'''
Created on 21 janv. 2021

@author: tleduc

Copyright 2020 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
from datetime import timezone
import warnings

from pysolar import solar
from suntimes.suntimes import SunTimes
from t4gpd.commons.sun.AbstractSunLib import AbstractSunLib

warnings.filterwarnings('ignore', message="I don't know about leap seconds after 2020")


class PySolarSunLib(AbstractSunLib):
    '''
    classdocs
    '''

    def __azimToTrigonometric(self, solarAzimuthAngle):
        return (360 - (solarAzimuthAngle - 90)) % 360

    def getNativeSolarAnglesInDegrees(self, dt):
        alti = solar.get_altitude(self.lat, self.lon, dt)
        azim = solar.get_azimuth(self.lat, self.lon, dt)
        return alti, azim

    def getSolarAnglesInDegrees(self, dt):
        alti, azim = self.getNativeSolarAnglesInDegrees(dt)
        azim = self.__azimToTrigonometric(azim)
        return alti, azim

    def getSolarDeclination(self, dt):
        _dayOfYear = self.getDayOfYear(dt)
        return solar.get_declination(_dayOfYear)

    def getSunrise(self, dt):
        return SunTimes(self.lon, self.lat, altitude=0).riseutc(dt).replace(tzinfo=timezone.utc)

    def getSunset(self, dt):
        return SunTimes(self.lon, self.lat, altitude=0).setutc(dt).replace(tzinfo=timezone.utc)



'''
Created on 31 janv. 2021

@author: tleduc

Copyright 2020 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
from geopandas.geodataframe import GeoDataFrame
from shapely.affinity import translate
from shapely.geometry import MultiPolygon, Polygon
from t4gpd.commons.GeomLib import GeomLib
from t4gpd.commons.IllegalArgumentTypeException import IllegalArgumentTypeException
from t4gpd.morph.geoProcesses.AbstractGeoprocess import AbstractGeoprocess


class FootprintExtruder(AbstractGeoprocess):
    '''
    classdocs
    '''

    def __init__(self, buildingsGdf, elevationFieldname='HAUTEUR', forceZCoordToZero=True,
                 withoutBuildingFloors=False):
        '''
        Constructor
        '''
        if not isinstance(buildingsGdf, GeoDataFrame):
            raise IllegalArgumentTypeException(buildingsGdf, 'GeoDataFrame')

        if not (elevationFieldname in buildingsGdf):
            raise Exception('%s is not a relevant field name!' % (elevationFieldname))
        self.elevationFieldname = elevationFieldname
        self.forceZCoordToZero = forceZCoordToZero
        self.withoutBuildingFloors = withoutBuildingFloors

    def __wallsPerRing(self, groundRing, roofRing):
        _down, _up = groundRing.coords, roofRing.coords
        _nNodes, result = len(_down), []
        for i in range(1, _nNodes):
            result.append(Polygon([_down[i - 1], _down[i], _up[i], _up[i - 1], _down[i - 1]]))
        return result

    def __walls(self, ground, roof):
        result = self.__wallsPerRing(ground.exterior, roof.exterior) 

        _nHoles = len(ground.interiors)
        for i in range(_nHoles):
            result += self.__wallsPerRing(ground.interiors[i], roof.interiors[i])

        return result

    def runWithArgs(self, row):
        _geom, _deltaZ = row.geometry, row[self.elevationFieldname]
        if not (isinstance(_geom, Polygon) or (isinstance(_geom, MultiPolygon) and (1 == len(_geom.geoms)))):
            raise IllegalArgumentTypeException(_geom, 'Polygon or MultiPolygon (with a single Polygon)')

        if isinstance(_geom, MultiPolygon):
            _geom = _geom.geoms[0]

        _geom = GeomLib.normalizeRingOrientation(_geom)

        if (self.forceZCoordToZero) or (not _geom.has_z):
            _ground = GeomLib.forceZCoordinateToZ0(_geom)
            _roof = translate(_ground, zoff=_deltaZ)
        else:
            _roof = _geom
            _ground = translate(_roof, zoff=-_deltaZ)

        _ground, _roof = (_ground, _roof) if (0 < _deltaZ) else (_roof, _ground)

        _walls = self.__walls(_ground, _roof)

        # TL. 10.03.2021
        _ground = GeomLib.normalizeRingOrientation(_ground, ccw=False)
        _roof = GeomLib.normalizeRingOrientation(_roof, ccw=True)

        if self.withoutBuildingFloors:
            return { 'geometry': MultiPolygon([_roof] + _walls) }
        return { 'geometry': MultiPolygon([_ground, _roof] + _walls) }



'''
Created on 7 avr. 2022

@author: tleduc

Copyright 2020-2022 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
from geopandas.geodataframe import GeoDataFrame
from t4gpd.commons.GeoProcess import GeoProcess
from t4gpd.commons.IllegalArgumentTypeException import IllegalArgumentTypeException
from t4gpd.commons.PointsDensifierLib import PointsDensifierLib


class STPointsDensifier2(GeoProcess):
    '''
    classdocs
    '''

    def __init__(self, gdf, curvAbsc, pathidFieldname=None, distToTheSubstrate=None):
        '''
        Constructor
        '''
        if not isinstance(gdf, GeoDataFrame):
            raise IllegalArgumentTypeException(gdf, 'GeoDataFrame')
        if not ((pathidFieldname is None) or (pathidFieldname in gdf)):
            raise Exception(f'{pathidFieldname} is not a valid fieldname!')

        self.gdf = gdf
        self.pathidFieldname = pathidFieldname
        self.curvAbsc = curvAbsc
        self.distToTheSubstrate = distToTheSubstrate

    def run(self):
        rows = []
        for rowId, row in self.gdf.iterrows():
            geom = row.geometry
            if not self.pathidFieldname is None:
                rowId = row[self.pathidFieldname]

            result = PointsDensifierLib.densifyByCurvilinearAbscissa(
                geom, self.curvAbsc, rowId, contourid=0,
                distToTheSubstrate=self.distToTheSubstrate)

            for _row in result:
                rows.append(self.updateOrAppend(row, _row))

        return GeoDataFrame(rows, crs=self.gdf.crs)



'''
Created on 17 juin 2020

@author: tleduc

Copyright 2020 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
from multiprocessing import Pool

from geopandas.geodataframe import GeoDataFrame
from shapely.geometry import LineString, Point, Polygon
from t4gpd.commons.GeoProcess import GeoProcess
from t4gpd.commons.IllegalArgumentTypeException import IllegalArgumentTypeException
from t4gpd.commons.RayCasting3Lib import RayCasting3Lib


class STIsovistField2D(GeoProcess):
    '''
    classdocs
    '''

    def __init__(self, buildings, viewpoints, nRays=64, rayLength=100.0, multiprocessing=False):
        '''
        Constructor
        '''
        if not isinstance(buildings, GeoDataFrame):
            raise IllegalArgumentTypeException(buildings, 'buildings GeoDataFrame')
        if not isinstance(viewpoints, GeoDataFrame):
            raise IllegalArgumentTypeException(viewpoints, 'viewpoints GeoDataFrame')
        if not (buildings.crs == viewpoints.crs):
            raise Exception('Illegal argument: buildings and viewpoints must share shames CRS!')

        self.buildings = buildings
        # CLEAN GEOMETRIES
        self.buildings.geometry = self.buildings.geometry.apply(lambda g: g.buffer(0))

        self.viewpoints = viewpoints
        self.shootingDirs = RayCasting3Lib.preparePanopticRays(nRays)
        self.rayLength = rayLength
        self.multiprocessing = multiprocessing

    @staticmethod
    def _process(params):
        row, masks, viewPoint, shootingDirs, rayLength = params
        return row, RayCasting3Lib.outdoorMultipleRayCast2D(masks, viewPoint, shootingDirs, rayLength)

    def run(self):
        rowsIsovRays, rowsIsov = list(), list()

        if self.multiprocessing:
            # PREPROCESS
            listOfParams = [
                (row, self.buildings, Point((row.geometry.x, row.geometry.y)), self.shootingDirs, self.rayLength)
                for _, row in self.viewpoints.iterrows()
                ]

            # PROCESS
            pool = Pool()
            quartets = pool.map(STIsovistField2D._process, listOfParams)
            pool.close()

            # POSTPROCESS
            for row, (rays, ctrPoints, _) in quartets:
                vp = Point((row.geometry.x, row.geometry.y))
                isov = Polygon(ctrPoints)
                isovCentre = isov.centroid
                if isovCentre.is_empty:
                    drift = None
                else:
                    drift = LineString([vp, isovCentre]).wkt

                rowsIsovRays.append(self.updateOrAppend(row, { 
                    'geometry': rays, 'viewpoint': vp.wkt, 'vect_drift': drift }))
                rowsIsov.append(self.updateOrAppend(row, {
                    'geometry': isov, 'viewpoint': vp.wkt, 'vect_drift': drift }))

        else:
            for _, row in self.viewpoints.iterrows():
                vp = Point((row.geometry.x, row.geometry.y))  # To handle Point Z
                rays, ctrPoints, _ = RayCasting3Lib.outdoorMultipleRayCast2D(self.buildings, vp, self.shootingDirs, self.rayLength)
                isov = Polygon(ctrPoints)
                isovCentre = isov.centroid
                if isovCentre.is_empty:
                    drift = None
                else:
                    drift = LineString([vp, isovCentre]).wkt

                rowsIsovRays.append(self.updateOrAppend(row, { 
                    'geometry': rays, 'viewpoint': vp.wkt, 'vect_drift': drift }))
                rowsIsov.append(self.updateOrAppend(row, {
                    'geometry': isov, 'viewpoint': vp.wkt, 'vect_drift': drift }))

        return [
            GeoDataFrame(rowsIsovRays, crs=self.viewpoints.crs),
            GeoDataFrame(rowsIsov, crs=self.viewpoints.crs)
            ]



'''
Created on 19 mars 2021

@author: tleduc

Copyright 2020 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
import geopandas
from geopandas.geodataframe import GeoDataFrame
from matplotlib.colors import ListedColormap
from pyvista import global_theme, Plotter
from shapely.affinity import scale, translate
from shapely.geometry.multipolygon import MultiPolygon
from shapely.ops import unary_union, triangulate
from shapely.wkt import loads
from t4gpd.commons.GeomLib import GeomLib
from t4gpd.morph.STPointsDensifier import STPointsDensifier
from t4gpd.morph.geoProcesses.FootprintExtruder import FootprintExtruder
from t4gpd.morph.geoProcesses.STGeoProcess import STGeoProcess
from t4gpd.pyvista.ToUnstructuredGrid import ToUnstructuredGrid

import matplotlib.pyplot as plt 


class Logos(object):
    '''
    classdocs
    '''

    @staticmethod
    def logoAsMultiLineString(_deltaX=50.0, _scaleX=0.9):
        _t = loads('MULTILINESTRING ((0 100, 0 0, 50 0), (0 50, 50 50))')

        _4 = loads('LINESTRING (50 50, 0 50, 30 100, 30 0)')
        _4 = translate(_4, xoff=_deltaX)

        _g = loads('LINESTRING (50 0, 0 0, 0 50, 50 50, 50 -50, 0 -50)')
        _g = translate(_g, xoff=2 * _deltaX)

        _p = loads('LINESTRING (0 -50, 0 50, 50 50, 50 0, 0 0)')
        _p = translate(_p, xoff=3 * _deltaX)

        _d = loads('LINESTRING (50 50, 0 50, 0 0, 50 0, 50 100)')
        _d = translate(_d, xoff=4 * _deltaX)

        _t4gpd = [_t, _4, _g, _p, _d]
        _t4gpd = [scale(letter, xfact=_scaleX) for letter in _t4gpd]
        _t4gpd = unary_union(_t4gpd)

        return _t4gpd

    @staticmethod
    def logoAsMultiPolygon(_deltaX=50.0, _scaleX=0.75, _buffDist=5.0):
        _mls = Logos.logoAsMultiLineString(_deltaX, _scaleX)
        return _mls.buffer(_buffDist)

    @staticmethod
    def logoAsMesh():
        gdf = GeoDataFrame([{'geometry':Logos.logoAsMultiPolygon()}])

        _gdf = STPointsDensifier(gdf, distance=45.0, pathidFieldname=None, adjustableDist=True, removeDuplicate=True).run()
        _tin = triangulate(unary_union(GeomLib.getListOfShapelyPoints(unary_union(_gdf.geometry))))
        _gdf = GeoDataFrame([{'geometry': MultiPolygon(_tin)}]).explode(ignore_index=True)
        gdf.geometry = gdf.geometry.apply(lambda g: g.buffer(1e-3))

        _gdf = geopandas.overlay(_gdf, gdf, how='intersection', keep_geom_type=True)
        _gdf['gid'] = range(len(_gdf))
        _gdf['gid'] = _gdf.gid % 4

        return _gdf

    @staticmethod
    def logoAsPdf(filename=None, greyscaled=False):
        _gdf = Logos.logoAsMesh()

        fig, _basemap = plt.subplots(figsize=(1 * 8.26, 1 * 8.26))

        if greyscaled:
            my_rgb_cmap = ListedColormap(['lightgrey', 'grey', 'darkgrey', 'dimgrey', 'black'])
            my_rgb_cmap = ListedColormap(['grey', 'darkgrey', 'dimgrey'])
            _gdf.plot(ax=_basemap, column='gid', edgecolor='grey', linewidth=0.1, cmap=my_rgb_cmap)
        else:
            _gdf.plot(ax=_basemap, color='dimgrey', edgecolor='lightgrey', linewidth=0.1)

        _basemap.axis('off')
        if filename is None:
            plt.show()
        else:
            fig.savefig(filename, bbox_inches='tight')

    @staticmethod
    def logo3DAsPdf(filename=None, greyscaled=False):
        _gdf = Logos.logoAsMesh()
        _gdf = _gdf.explode(ignore_index=True)
        _gdf['HAUTEUR'] = 15.0

        op = FootprintExtruder(_gdf, elevationFieldname='HAUTEUR', forceZCoordToZero=True)
        _gdf = STGeoProcess(op, _gdf).run()

        global_theme.background = 'white'
        global_theme.axes.show = False
        global_theme.lighting = True
        global_theme.camera = {
            'position': [859.078, -4593.389, 6933.603],
            'focal_point': [4287.524, 3160.188, 87.055],
            'viewup': [0.254, 0.575, 0.778]
            }
        if filename is None:
            plotter = Plotter(window_size=(1000, 800))
        else:
            plotter = Plotter(off_screen=True, window_size=(1000, 800))

        if greyscaled:
            _logo3D = ToUnstructuredGrid([_gdf], fieldname=None).run()
            plotter.add_mesh(_logo3D, color='dimgrey',
                             show_edges=False, show_scalar_bar=False)
        else:
            _logo3D = ToUnstructuredGrid([_gdf], fieldname='gid').run()
            # my_colormap = ListedColormap(['red', 'cyan', 'yellow', 'dimgrey'])
            my_colormap = 'Pastel1'
            plotter.add_mesh(_logo3D, scalars='gid', cmap=my_colormap,
                             show_edges=False, show_scalar_bar=False)
        plotter.camera.zoom(1.4)
        plotter.show(screenshot=filename)
        return _gdf



'''
Created on 21 janv. 2021

@author: tleduc

Copyright 2020 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
from datetime import datetime, timedelta, timezone

from numpy import arccos, arctan2, cos, pi, sin, tan
from t4gpd.commons.AngleLib import AngleLib

from t4gpd.commons.sun.AbstractSunLib import AbstractSunLib


class SoleneSunLib(AbstractSunLib):
    '''
    classdocs
    '''
    MINS1440 = 1440
    PI_DIV_2 = pi / 2.0
    PI_TIMES_2 = pi * 2.0

    def __fromHourAngleToMinutes(self, hourAngle):
        return int(720.0 + AngleLib.toDegrees(hourAngle) / 0.25)

    @staticmethod
    def __getEquationOfTime(fractionalYear):
        '''
        Based on "General Solar Position Calculations - NOAA Global Monitoring Division"
        https://www.esrl.noaa.gov/gmd/grad/solcalc/solareqns.PDF
        '''
        return 229.18 * (0.000075 + 
                         0.001868 * cos(fractionalYear) - 
                         0.032077 * sin(fractionalYear) - 
                         0.014615 * cos(2 * fractionalYear) - 
                         0.040849 * sin(2 * fractionalYear))

    def __getFractionalYear(self, dt):
        '''
        Based on "General Solar Position Calculations - NOAA Global Monitoring Division"
        https://www.esrl.noaa.gov/gmd/grad/solcalc/solareqns.PDF
        '''
        _dayOfYear = self.getDayOfYear(dt)
        return 2 * pi * (_dayOfYear - 1.0 + (dt.hour - 12.0) / 24.0) / self.nDaysPerYear(dt)

    def getSolarAnglesInDegrees(self, dt):
        _tz = 0 if (dt.utcoffset() is None) else dt.utcoffset().seconds / 3600
        _latInRadians = AngleLib.toRadians(self.lat)

        _fractionalYear = self.__getFractionalYear(dt)
        _eqTime = SoleneSunLib.__getEquationOfTime(_fractionalYear)
        _declination = self.__getSolarDeclinationInRadians(_fractionalYear)

        hour, minute = dt.hour, dt.minute
        # timeOffset = _eqTime + 4 * self.lon - 60 * _tz
        # According to: http://www.jgiesen.de/astro/suncalc/calculations.htm
        timeOffset = _eqTime - 4 * self.lon + 60 * _tz

        # trueSolarTime = hour * 60.0 + minute + second / 60.0 + timeOffset
        # According to: http://www.jgiesen.de/astro/suncalc/calculations.htm
        trueSolarTime = hour * 60.0 + minute + timeOffset
        '''
        Observing the sun from earth, the solar hour angle is an expression of
        time, expressed in angular measurement, usually degrees, from solar noon.
        At solar noon the hour angle is 0.0 degree. The time before solar noon is
        expressed as negative degrees. 24hrs correspond to 360 degrees, 15
        degrees per hour or 0.25 degree per minute.
        '''
        # solarHourAngle = AngleLib.toRadians((trueSolarTime - 12.0 * 60) * 0.25)
        # According to: http://www.jgiesen.de/astro/suncalc/calculations.htm
        solarHourAngle = AngleLib.toRadians((trueSolarTime / 4.0) - 180)

        cosLat, sinLat = cos(_latInRadians), sin(_latInRadians)
        cosDeclination, sinDeclination = [cos(_declination), sin(_declination)]
        cosHourAngle = cos(solarHourAngle)
        cosDeclinationCosHourAngle = cosDeclination * cosHourAngle

        cosSolarZenithAngle = (sinLat * sinDeclination + cosLat * cosDeclinationCosHourAngle)

        solarZenithAngle = arccos(cosSolarZenithAngle)
        alti = SoleneSunLib.PI_DIV_2 - solarZenithAngle

        # According to: https://researchlibrary.agric.wa.gov.au/cgi/viewcontent.cgi?article=1122&context=rmtr
        azim = SoleneSunLib.PI_DIV_2 - arctan2(
            -sin(solarHourAngle), (tan(_declination) * cos(_latInRadians) - sin(_latInRadians) * cos(solarHourAngle))
            )
        if (0 > azim):
            azim += SoleneSunLib.PI_TIMES_2

        return AngleLib.toDegrees(alti), AngleLib.toDegrees(azim)

    def __getSolarDeclinationInRadians(self, fractionalYear):
        '''
        Based on "General Solar Position Calculations - NOAA Global Monitoring Division"
        https://www.esrl.noaa.gov/gmd/grad/solcalc/solareqns.PDF

        Solar Declination by Spencer (1971)
        '''
        return (0.006918 - 
                0.399912 * cos(fractionalYear) + 
                0.070257 * sin(fractionalYear) - 
                0.006758 * cos(2 * fractionalYear) + 
                0.000907 * sin(2 * fractionalYear) - 
                0.002697 * cos(3 * fractionalYear) + 
                0.00148 * sin(3 * fractionalYear))

    def getSolarDeclination(self, dt):
        _fractionalYear = self.__getFractionalYear(dt)
        _declination = self.__getSolarDeclinationInRadians(_fractionalYear)
        return AngleLib.toDegrees(_declination)

    def getSunrise(self, dt):
        _latInRadians = AngleLib.toRadians(self.lat)
        _fractionalYear = self.__getFractionalYear(dt)
        _declination = self.__getSolarDeclinationInRadians(_fractionalYear)
        '''
        _solarHourAngle = arccos(
            cos(AngleLib.toRadians(90.833)) / (cos(_latInRadians) * cos(_declination)) - 
            tan(_latInRadians) * tan(_declination))
        return 720 - 4 * (self.lon + _solarHourAngle) - _eqTime
        '''
        _solarHourAngle = -arccos(-tan(_declination) * tan(_latInRadians))
        _delta = self.__fromHourAngleToMinutes(_solarHourAngle)
        _today = datetime(dt.year, dt.month, dt.day, tzinfo=timezone.utc)
        return _today + timedelta(minutes=_delta)

    def getSunset(self, dt):
        _latInRadians = AngleLib.toRadians(self.lat)
        _fractionalYear = self.__getFractionalYear(dt)
        _declination = self.__getSolarDeclinationInRadians(_fractionalYear)
        '''
        _solarHourAngle = arccos(
            cos(AngleLib.toRadians(90.833)) / (cos(_latInRadians) * cos(_declination)) - 
            tan(_latInRadians) * tan(_declination))
        return 720 - 4 * (self.lon + _solarHourAngle) - _eqTime
        '''
        _solarHourAngle = -arccos(-tan(_declination) * tan(_latInRadians))
        _delta = self.MINS1440 - self.__fromHourAngleToMinutes(_solarHourAngle)
        _today = datetime(dt.year, dt.month, dt.day, tzinfo=timezone.utc)
        return _today + timedelta(minutes=_delta)



'''
Created on 17 feb. 2022

@author: tleduc

Copyright 2020-2022 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''

from geopandas import GeoDataFrame
from pandas import read_csv
from shapely.wkt import loads


class AbstractGeoDataFrameDemos(object):
    '''
    classdocs
    '''

    @staticmethod
    def postprocess(sio, crs='epsg:2154'):
        _df = read_csv(sio, sep=';')
        _df.rename(columns={'Unnamed: 0': ''}, inplace=True)
        _df.set_index('', inplace=True)
        _df.reset_index(drop=True, inplace=True)
        _df.geometry = _df.geometry.apply(lambda g: loads(g))
        return GeoDataFrame(_df, crs=crs)

    @staticmethod
    def postprocess2(sio, crs='epsg:2154'):
        _df = read_csv(sio, sep=';')
        _df.geometry = _df.geometry.apply(lambda g: loads(g))
        return GeoDataFrame(_df, crs=crs)



'''
Created on 12 nov. 2020

@author: tleduc

Copyright 2020 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
from geopandas.geodataframe import GeoDataFrame
from shapely.geometry import Point

from t4gpd.commons.GeoProcess import GeoProcess
from t4gpd.commons.IllegalArgumentTypeException import IllegalArgumentTypeException
from t4gpd.commons.graph.ShortestPathLib import ShortestPathLib


class STShortestPath(GeoProcess):
    '''
    classdocs
    '''

    def __init__(self, roads, fromPoints, toPoints):
        '''
        Constructor
        '''
        if not isinstance(roads, GeoDataFrame):
            raise IllegalArgumentTypeException(roads, 'GeoDataFrame')
        if not isinstance(fromPoints, (GeoDataFrame, Point)):
            raise IllegalArgumentTypeException(fromPoints, 'GeoDataFrame or single Point')
        if not isinstance(toPoints, (GeoDataFrame, Point)):
            raise IllegalArgumentTypeException(toPoints, 'GeoDataFrame or single Point')

        self.graph = ShortestPathLib(roads)
        self.roads = roads

        if isinstance(fromPoints, GeoDataFrame):
            self.fromPoints = list(fromPoints.centroid)
        else:
            self.fromPoints = [fromPoints]

        if isinstance(toPoints, GeoDataFrame):
            self.toPoints = list(toPoints.centroid)
        else:
            self.toPoints = [toPoints]

    def run(self):
        gid, rows = 0, []
        for fromPoint in self.fromPoints:
            for toPoint in self.toPoints:
                pathGeom, pathLen = self.graph.shortestPath(fromPoint, toPoint)
                if not pathGeom is None:
                    rows.append({'gid':gid, 'pathLen': pathLen, 'fromPoint': fromPoint.wkt,
                                 'toPoint': toPoint.wkt, 'geometry': pathGeom})
                    gid += 1

        return GeoDataFrame(rows, crs=self.roads.crs)



'''
Created on 5 avr. 2022

@author: tleduc

Copyright 2020-2022 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''

from numpy import arange, array, ndarray
from pyvista import MultiBlock, UnstructuredGrid
from t4gpd.commons.GeoProcess import GeoProcess
from t4gpd.commons.GeomLib import GeomLib
from t4gpd.commons.IllegalArgumentTypeException import IllegalArgumentTypeException
from vtkmodules.util.vtkConstants import VTK_POLYGON, VTK_POLY_LINE, VTK_TRIANGLE, \
    VTK_VERTEX


class ToUnstructuredGrid(GeoProcess):
    '''
    classdocs
    '''

    def __init__(self, gdfs, fieldname=None):
        '''
        Constructor
        '''
        if not isinstance(gdfs, (list, tuple, ndarray)):
            raise IllegalArgumentTypeException(gdfs, 'list or tuple of GeoDataFrames')
        self.gdfs = gdfs
        self.fieldname = fieldname

    def __fromGeomToUnstructuredGrids(self, geom):
        # https://vtk.org/wp-content/uploads/2015/04/file-formats.pdf
        if ('Point' == geom.geom_type):
            point = [[xyz[0], xyz[1], 0 if len(xyz) == 2 else xyz[2]] for xyz in geom.coords]
            cells = [1, 0]
            cells, celltypes, points = array(cells), array([VTK_VERTEX]), array(point)
            return [ UnstructuredGrid(cells, celltypes, points) ]

        elif ('LineString' == geom.geom_type):
            points = [[xyz[0], xyz[1], 0 if len(xyz) == 2 else xyz[2]] for xyz in geom.coords]
            npoints = len(points)
            cells = [npoints] + arange(npoints).tolist()
            cells, celltypes, points = array(cells), array([VTK_POLY_LINE]), array(points)
            return [ UnstructuredGrid(cells, celltypes, points) ]

        elif ('Polygon' == geom.geom_type):
            if (GeomLib.isConvex(geom) or GeomLib.isHoled(geom)):
                points = [[xyz[0], xyz[1], 0 if len(xyz) == 2 else xyz[2]] for xyz in geom.exterior.coords]
                npoints = len(points)
                cells = [npoints] + arange(npoints).tolist()
                cells, celltypes, points = array(cells), array([VTK_POLYGON]), array(points)
                return [ UnstructuredGrid(cells, celltypes, points) ]
            else:
                result = []
                triangles = GeomLib.fromPolygonToListOfTriangles(geom)
                for triangle in triangles:
                    points = [[xyz[0], xyz[1], 0 if len(xyz) == 2 else xyz[2]] for xyz in triangle.exterior.coords]
                    npoints = len(points)
                    cells = [npoints] + arange(npoints).tolist()
                    cells, celltypes, points = array(cells), array([VTK_TRIANGLE]), array(points)
                    result.append(UnstructuredGrid(cells, celltypes, points))
                return result

        raise Exception('Unreachable instruction!')

    def __fromGeoDataFrameToDictOfUnstructuredGrids(self, gdf, rank):
        gdf = gdf.explode(ignore_index=True)
        # print(gdf.columns)
        result = {}
        for idx, row in gdf.iterrows():
            geom = row.geometry
            partialPolyUgrids = self.__fromGeomToUnstructuredGrids(geom)

            # we can add some attributes to the current UnstructuredGrid
            if self.fieldname is not None:
                attrValue = row[self.fieldname] if self.fieldname in gdf else float('nan')
                for partialPolyUgrid in partialPolyUgrids:
                    partialPolyUgrid.cell_data[self.fieldname] = attrValue

            for i, partialPolyUgrid in enumerate(partialPolyUgrids):
                result[f'{rank}_{idx}_{i}'] = partialPolyUgrid
        return result

    def run(self):
        result = {}
        rank = 0
        for rank, gdf in enumerate(self.gdfs):
            result.update(self.__fromGeoDataFrameToDictOfUnstructuredGrids(gdf, rank))
        polyBlocks = MultiBlock(result)
        polyGrid = polyBlocks.combine()
        return polyGrid



'''
Created on 10 sept. 2021

@author: tleduc

Copyright 2020-2021 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
from qgis.core import (QgsArrowSymbolLayer, QgsCategorizedSymbolRenderer,
                       QgsFillSymbol, QgsHeatmapRenderer,
                       QgsLineSymbol, QgsMarkerSymbol,
                       QgsPalLayerSettings, QgsLinePatternFillSymbolLayer,
                       QgsPointPatternFillSymbolLayer, QgsProperty,
                       QgsRendererCategory, QgsSimpleFillSymbolLayer,
                       QgsSimpleLineSymbolLayer, QgsStyle,
                       QgsSvgMarkerSymbolLayer, QgsSymbol, QgsSymbolLayer,
                       QgsTextFormat, QgsVectorLayerSimpleLabeling)
from qgis.PyQt.QtGui import QColor
from qgis.utils import iface


class SetSymbolLib(object):
    '''
    classdocs
    '''

    @staticmethod
    def __setSymbol(layer, symbol):
        if layer is not None:
            layer.renderer().setSymbol(symbol)
            iface.layerTreeView().refreshLayerSymbology(layer.id())
            layer.triggerRepaint()

    @staticmethod
    def __altSetSymbol(layer, symbol):
        if layer is not None:
            layer.renderer().symbol().changeSymbolLayer(0, symbol)
            iface.layerTreeView().refreshLayerSymbology(layer.id())
            layer.triggerRepaint()

    @staticmethod
    def __otherAltSetSymbol(layer, renderer):
        if layer is not None:
            layer.setRenderer(renderer)
            iface.layerTreeView().refreshLayerSymbology(layer.id())
            layer.triggerRepaint()

    @staticmethod
    def setAlternateFillSymbol(layer, color='black', patternDistance=1.0,
                               lineAngle=45, penstyle='solid', width='0.35'):
        lineSymbol = QgsLineSymbol.createSimple({
            'penstyle': penstyle, 'width': width, 'color': color})
        filled_pattern = QgsLinePatternFillSymbolLayer()
        filled_pattern.setLineAngle(lineAngle)
        filled_pattern.setDistance(patternDistance)
        filled_pattern.setSubSymbol(lineSymbol)
        SetSymbolLib.__altSetSymbol(layer, filled_pattern)

    @staticmethod
    def setAlternateFillSymbol2(layer, color='red', color_border='black',
                                name='diamond', size='3.0'):
        markerSymbol = QgsMarkerSymbol.createSimple({
            'color':color, 'color_border':color_border, 'name':name, 'size':size})
        filled_pattern = QgsPointPatternFillSymbolLayer()
        filled_pattern.setDistanceX(4.0)
        filled_pattern.setDistanceY(4.0)
        filled_pattern.setSubSymbol(markerSymbol)
        SetSymbolLib.__altSetSymbol(layer, filled_pattern)

    @staticmethod
    def setAlternateFillSymbol3(layer, color='black', style='b_diagonal'):
        # style = ['cross', 'b_diagonal', 'diagonal_x', 'f_diagonal', 
        # 'horizontal', 'solid', 'vertical']
        symbol_layer = QgsSimpleFillSymbolLayer.create({'color':color, 'style': style})
        SetSymbolLib.__altSetSymbol(layer, symbol_layer)

    @staticmethod
    def setArrowSymbol(layer, color='black', headType=QgsArrowSymbolLayer.HeadSingle,
                       width=0.60, headLength=2.05, headThickness=1.55,
                       arrowType=QgsArrowSymbolLayer.ArrowPlain):
        # headType = {QgsArrowSymbolLayer.HeadDouble, QgsArrowSymbolLayer.HeadReversed, QgsArrowSymbolLayer.HeadSingle}
        # arrowType = {QgsArrowSymbolLayer.ArrowPlain, QgsArrowSymbolLayer.ArrowLeftHalf, QgsArrowSymbolLayer.ArrowRightHalf}

        arrowSymbol = QgsArrowSymbolLayer()
        arrowSymbol.setColor(QColor(color))

        arrowSymbol.setArrowStartWidth(width)
        arrowSymbol.setArrowWidth(width)

        arrowSymbol.setArrowType(arrowType)

        arrowSymbol.setHeadType(headType)
        arrowSymbol.setHeadLength(headLength)
        arrowSymbol.setHeadThickness(headThickness)

        arrowSymbol.setIsCurved(False)
        arrowSymbol.setIsRepeated(True)

        SetSymbolLib.__altSetSymbol(layer, arrowSymbol)

    @staticmethod
    def setCategorizedSymbol(layer, fieldName, listOfValueColorLabels):
        '''
        listOfValueColorLabels = (
            ('cat', 'red'),
            ('dog', 'blue', 'Big dog'),
            ('sheep', 'green'),
            ('', 'yellow', 'Unknown')
        )
        '''
        # CREATE A CATEGORY FOR EACH ITEM IN 'fieldLookupTable'
        categories = []
        for item in listOfValueColorLabels:
            value, color, label = item if 3 == len(item) else [item[0], item[1], item[0]]
            symbol = QgsSymbol.defaultSymbol(layer.geometryType())
            symbol.setColor(QColor(color))
            categories.append(QgsRendererCategory(value, symbol, label))

        # CREATE THE RENDERER AND ASSIGN IT TO THE GIVEN LAYER
        renderer = QgsCategorizedSymbolRenderer(fieldName, categories)
        SetSymbolLib.__otherAltSetSymbol(layer, renderer)

    @staticmethod
    def setFillSymbol(layer, color='lightgrey', outline_color='darkgrey',
                      style_border='solid', width_border='0.75'):
        fillSymbol = QgsFillSymbol.createSimple({
            'color': color,
            'outline_color': outline_color,
            'width_border': width_border,
            'style_border': style_border})
        SetSymbolLib.__setSymbol(layer, fillSymbol)

    @staticmethod
    def setHeatMap(layer, radius=20, rampName='Blues'):
        # To get a list of available color ramp names, use:
        # QgsStyle().defaultStyle().colorRampNames()
        ramp = QgsStyle().defaultStyle().colorRamp(rampName)
        heatmap = QgsHeatmapRenderer()
        heatmap.setRadius(radius)
        heatmap.setColorRamp(ramp)
        SetSymbolLib.__otherAltSetSymbol(layer, heatmap)

    @staticmethod
    def setLabelSymbol(layer, fieldName, overPoint=True, size='14',
                       color='black', positionX=None, positionY=None,
                       offsetX=None, offsetY=None):
        label = QgsPalLayerSettings()
        label.fieldName = fieldName
    
        textFormat = QgsTextFormat()
        # bgColor = QgsTextBackgroundSettings()
        # bgColor.setFillColor(QColor('white'))
        # bgColor.setEnabled(True)
        # textFormat.setBackground(bgColor )
        textFormat.setColor(QColor(color))
        textFormat.setSize(int(size))
        label.setFormat(textFormat)
        label.isExpression = True
    
        layer.setLabeling(QgsVectorLayerSimpleLabeling(label))
        layer.setLabelsEnabled(True)
        iface.layerTreeView().refreshLayerSymbology(layer.id())
        layer.triggerRepaint()

    @staticmethod
    def setLineSymbol(layer, color='black', penstyle='solid', width='0.55'):
        # dash, dash dot, dash dot dot, dot, solid
        lineSymbol = QgsLineSymbol.createSimple({
            'color': color, 'penstyle': penstyle, 'width': width})
        SetSymbolLib.__setSymbol(layer, lineSymbol)

    @staticmethod
    def setMarkerSymbol(layer, color='black', size='3.6', name='circle'):
        # circle, square, rectangle, diamond, pentagon, triangle, 
        # equilateral_triangle, star, regular_star, arrow 
        nodesSymbol = QgsMarkerSymbol.createSimple({
            'color': color, 'name': name, 'size': size, 'width_border': '0'})
        SetSymbolLib.__setSymbol(layer, nodesSymbol)

    @staticmethod
    def setSimpleOutlineFillSymbol(layer, color='red', width='1.1', penstyle='dot'):
        symbol_layer = QgsSimpleLineSymbolLayer.create(
            {'color': color, 'width': width, 'penstyle': penstyle})
        SetSymbolLib.__altSetSymbol(layer, symbol_layer)

    @staticmethod
    def setSvgSymbol(markerLayer, fieldName, svgSymbolDirname,
                     size='6', rotationFieldName=None):
        if markerLayer is not None:
            fni = markerLayer.dataProvider().fieldNameIndex(fieldName)
            uniqValues = markerLayer.dataProvider().uniqueValues(fni)

            categories = []
            for value in uniqValues:
                mySymbol = QgsSymbol.defaultSymbol(markerLayer.geometryType())

                svgStyle = { 'name': f'{svgSymbolDirname}/{value}.svg', 'size': size }
                svgSymbol = QgsSvgMarkerSymbolLayer.create(svgStyle)
                if rotationFieldName is not None:
                    svgSymbol.dataDefinedProperties().setProperty(
                        QgsSymbolLayer.PropertyAngle, QgsProperty.fromField(rotationFieldName))

                mySymbol.changeSymbolLayer(0, svgSymbol)

                category = QgsRendererCategory(value, mySymbol, str(value))
                categories.append(category)

            renderer = QgsCategorizedSymbolRenderer(fieldName, categories)
            markerLayer.setRenderer(renderer)
            iface.layerTreeView().refreshLayerSymbology(markerLayer.id())
            markerLayer.triggerRepaint()



'''
Created on 31 aug. 2020

@author: tleduc

Copyright 2020 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
from geopandas.geodataframe import GeoDataFrame
from t4gpd.commons.GeoProcess import GeoProcess
from t4gpd.commons.IllegalArgumentTypeException import IllegalArgumentTypeException


class CSVWKTWriter(GeoProcess):
    '''
    classdocs
    '''

    def __init__(self, inputGdf, outputFile, fieldSep=';', decimalSep='.'):
        '''
        Constructor
        '''
        if not isinstance(inputGdf, GeoDataFrame):
            raise IllegalArgumentTypeException(inputGdf, 'GeoDataFrame')
        self.inputGdf = inputGdf
        self.outputFile = outputFile
        self.fieldSep = fieldSep
        self.decimalSep = decimalSep

    def __getFields(self):
        fields = dict()
        for fname in self.inputGdf.columns:
            ftype = str(self.inputGdf.dtypes[fname])
            if 'geometry' != ftype:
                fields[fname] = ftype
        return fields

    def __printFieldValue(self, fields, row, fname):
        if ('.' != self.decimalSep) and ('float64' == fields[fname]):
            return str(row[fname]).replace('.', self.decimalSep)
        return str(row[fname])

    def run(self):
        fields = self.__getFields()

        with open(self.outputFile, 'w') as f:
            f.write('%s%sthe_geom\n' % (
                self.fieldSep.join([fname for fname in fields.keys()]),
                self.fieldSep
                ))

            for _, row in self.inputGdf.iterrows():
                wktGeom = row.geometry.wkt
                tmp = self.fieldSep.join([self.__printFieldValue(fields, row, fname) for fname in fields.keys()] + [wktGeom])
                f.write(tmp + '\n')



'''
Created on 12 mai 2021

@author: tleduc

Copyright 2020-2021 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
from t4gpd.comfort.algo.ConstantsLib import ConstantsLib
from t4gpd.comfort.algo.PMVLib import PMVLib


class PTLib(object):
    '''
    classdocs
    '''

    @staticmethod
    def assess_pt(AirTC, RH, WS_ms, T_mrt):
        PMV = PMVLib.assess_pmv(AirTC, RH, WS_ms, T_mrt)

        # PT: Perceived Temperature
        # PT calculation based on clothing level, after (Coccolo et al., 2016)
        # The computation is based on the PMV for a standard pedestrian, called 
        # "Clima Michel", walking (135 W.m-2) and changing clothing according to 
        # the season:winter clothing (Icl equal to 1.75 clo) and summer clothing 
        # (Icl equal to 0.5 clo).
        PT = None

        if ((0 > PMV) and (ConstantsLib.Clo == 1.75)):
            PT = 5.805 + 12.6784 * PMV
        elif ((0 == PMV) and (0.5 < ConstantsLib.Clo < 1.75)):
            PT = 21.258 - 9.558 * ConstantsLib.Clo
        elif ((0 < PMV) and (0.5 == ConstantsLib.Clo)):
            PT = 16.826 + 6.183 * PMV

        return PT



'''
Created on 29 juin 2020

@author: tleduc

Copyright 2020 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
from geopandas.geodataframe import GeoDataFrame
from t4gpd.commons.BoundingBox import BoundingBox
from t4gpd.commons.GeoProcess import GeoProcess
from t4gpd.commons.IllegalArgumentTypeException import IllegalArgumentTypeException


class STClip(GeoProcess):
    '''
    classdocs
    '''

    def __init__(self, inputGdf, roi):
        '''
        Constructor
        '''
        if not isinstance(inputGdf, GeoDataFrame):
            raise IllegalArgumentTypeException(inputGdf, 'GeoDataFrame')
        self.inputGdf = inputGdf
        self.roi = BoundingBox(roi).asPolygon()
        
    def run(self):
        spatialIdx = self.inputGdf.sindex
        ids = spatialIdx.intersection(self.roi.bounds)
        rows = []

        for _id in ids:
            row = self.inputGdf.loc[_id]
            geom = row.geometry
            if self.roi.intersects(geom):
                g = self.roi.intersection(geom)
                rows.append(self.updateOrAppend(row, { 'geometry':g }))

        return GeoDataFrame(rows, crs=self.inputGdf.crs)



'''
Created on 22 juin 2022

@author: tleduc

Copyright 2020-2022 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
from geopandas import GeoDataFrame
from numpy import unique
from pandas import DataFrame
from t4gpd.commons.GeoProcess import GeoProcess


class STRaysToViewFactors(GeoProcess):
    '''
    classdocs
    '''

    def __init__(self, raysGdf, pkFieldname, hitGidsFieldname):
        '''
        Constructor
        '''
        assert isinstance(raysGdf, GeoDataFrame), 'raysGdf must be a GeoDataFrame'
        assert (pkFieldname in raysGdf), f'{pkFieldname} must be a GeoDataFrame field name!'
        assert (hitGidsFieldname in raysGdf), f'{hitGidsFieldname} must be a GeoDataFrame field name!'
        self.raysGdf = raysGdf
        self.pkFieldname = pkFieldname
        self.hitGidsFieldname = hitGidsFieldname

    def run(self):
        rows = []
        for _, row in self.raysGdf.iterrows():
            pk = row[self.pkFieldname]
            _hitGids = row[self.hitGidsFieldname]
            nrays = len(_hitGids)
            hitGids = [gid for gid in _hitGids if (gid is not None)]
            hitGids, hitGidsCount = unique(hitGids, return_counts=True)
            
            for i in range(len(hitGids)):
                rows.append({
                    'src': pk,
                    'dst': hitGids[i],
                    'viewfactor': hitGidsCount[i] / nrays,
                    }) 
        return DataFrame(data=rows)



'''
Created on 22 nov. 2020

@author: tleduc

Copyright 2020 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
import copy

from geopandas.geodataframe import GeoDataFrame
from shapely.geometry import Point

from t4gpd.commons.GeomLib import GeomLib
from t4gpd.commons.graph.AbstractUrbanGraphLib import AbstractUrbanGraphLib
from t4gpd.commons.graph.ShortestPathLib import ShortestPathLib


class NeighborhoodLib(AbstractUrbanGraphLib):
    '''
    classdocs
    '''

    @staticmethod
    def neighborhood(roads, fromPoint, maxDist):
        roi = fromPoint.buffer(maxDist, 1).bounds
        mygraph = ShortestPathLib(roads, roi)

        str_coord = AbstractUrbanGraphLib.hashCoord([fromPoint.x, fromPoint.y])
        if not str_coord in mygraph.ciVertices:
            _, nearestLine = AbstractUrbanGraphLib.addEndPoint(
                roads, roads.sindex, mygraph.graph, mygraph.ciVertices, mygraph.icVertices, mygraph.edges, fromPoint)

            _foo = lambda coord: mygraph.ciVertices[AbstractUrbanGraphLib.hashCoord(coord)]
            first, last = _foo(nearestLine.coords[0]), _foo(nearestLine.coords[-1])
            if (first in mygraph.edges) and (last in mygraph.edges[first]):
                del(mygraph.edges[first][last])
            if (last in mygraph.edges) and (first in mygraph.edges[last]):
                del(mygraph.edges[last][first])

        _icVertices = copy.deepcopy(mygraph.icVertices)

        # Identification of vertices in the neighborhood
        candidates = dict()
        for _vtxIdx, _vtxGeom in _icVertices.items():
            _geom = Point(_vtxGeom)
            if (maxDist >= fromPoint.distance(_geom)):
                pathGeom, pathLen = mygraph.shortestPath(fromPoint, _geom)
                if (pathGeom is not None) and (maxDist >= pathLen):
                    candidates[_vtxIdx] = maxDist - pathLen

        # Identification of edges in the neighborhood
        _burnedEdges, rows, gid = dict(), [], 0
        for _vtxIdx in candidates.keys():
            _remainingDist = candidates[_vtxIdx]
            for _neighborIdx in mygraph.edges[_vtxIdx].keys():
                _hash = '%d_%d' % (min(_neighborIdx, _vtxIdx), max(_neighborIdx, _vtxIdx))
                if _hash not in _burnedEdges:
                    _edgeItem = mygraph.edges[_vtxIdx][_neighborIdx]
                    if (_remainingDist >= _edgeItem['length']):
                        rows.append({ 'gid': gid, 'geometry': _edgeItem['geometry'], 'length': _edgeItem['length']})
                        _burnedEdges[_hash] = None
                        gid += 1

                    elif (_neighborIdx in candidates):
                        if ((candidates[_neighborIdx] >= _edgeItem['length']) or
                            (_remainingDist + candidates[_neighborIdx] >= _edgeItem['length'])):
                            rows.append({ 'gid': gid, 'geometry': _edgeItem['geometry'], 'length': _edgeItem['length']})
                            _burnedEdges[_hash] = None
                            gid += 1
                        else:
                            tmp = GeomLib.cutsLineStringByCurvilinearDistance(_edgeItem['geometry'], _remainingDist)
                            rows.append({ 'gid': gid, 'geometry': tmp[0], 'length': _remainingDist})
                            _burnedEdges[_hash] = None
                            gid += 1

                            _otherEdgeItem = mygraph.edges[_neighborIdx][_vtxIdx]
                            tmp = GeomLib.cutsLineStringByCurvilinearDistance(_otherEdgeItem['geometry'], candidates[_neighborIdx])
                            rows.append({ 'gid': gid, 'geometry': tmp[0], 'length': candidates[_neighborIdx]})
                            gid += 1

                    else:
                        tmp = GeomLib.cutsLineStringByCurvilinearDistance(_edgeItem['geometry'], _remainingDist)
                        rows.append({ 'gid': gid, 'geometry': tmp[0], 'length': _remainingDist})
                        _burnedEdges[_hash] = None
                        gid += 1
        '''
        rows2 = []
        for vtxIdx in candidates.keys():
            rows2.append({'gid':vtxIdx, 'remain_dist': candidates[vtxIdx], 'geometry': Point(mygraph.icVertices[vtxIdx])})
        return GeoDataFrame(rows, crs=roads.crs), GeoDataFrame(rows2, crs=roads.crs)
        '''
        return GeoDataFrame(rows, crs=roads.crs)



'''
Created on 9 avr. 2021

@author: tleduc

Copyright 2020-2021 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
from geopandas.geodataframe import GeoDataFrame
from t4gpd.commons.GeomLib import GeomLib
from t4gpd.commons.IllegalArgumentTypeException import IllegalArgumentTypeException


class AbstractGridLib(object):
    '''
    classdocs
    '''

    def __init__(self, gdf, dx, dy=None, encode=True):
        '''
        Constructor
        '''
        if not isinstance(gdf, GeoDataFrame):
            raise IllegalArgumentTypeException(gdf, 'GeoDataFrame')
        self.gdf = gdf
        self.dx = dx
        self.dy = dx if (dy is None) else dy
        self.encode = encode

    def grid(self):
        raise NotImplementedError('grid() must be overridden!')

    def indoorGrid(self):
        _grid = self.indoorOutdoorGrid()
        _grid = _grid.loc[ _grid[_grid.indoor == 1].index ]
        _grid.reset_index(drop=True, inplace=True)
        return _grid

    def indoorOutdoorGrid(self):
        _grid = self.grid()
        sidx = self.gdf.sindex
        rows = []
        for _, row in _grid.iterrows():
            _row = dict(row)
            _centroid = row['geometry'].centroid
            _indoor = 1 if (GeomLib.isAnIndoorPoint(_centroid, self.gdf, sidx)) else 0
            _row['indoor'] = _indoor
            rows.append(_row)
        _grid = GeoDataFrame(rows, crs=self.gdf.crs)
        return _grid

    def outdoorGrid(self):
        _grid = self.indoorOutdoorGrid()
        _grid = _grid.loc[ _grid[_grid.indoor == 0].index ]
        _grid.reset_index(drop=True, inplace=True)
        return _grid



'''
Created on 12 sept. 2020

@author: tleduc

Copyright 2020 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
from geopandas.geodataframe import GeoDataFrame
from shapely.ops import polygonize, unary_union
from t4gpd.commons.GeoProcess import GeoProcess
from t4gpd.commons.IllegalArgumentTypeException import IllegalArgumentTypeException
from t4gpd.commons.GeomLib import GeomLib
from t4gpd.commons.ArrayCoding import ArrayCoding


class STMultipleOverlaps(GeoProcess):
    '''
    classdocs
    '''

    def __init__(self, inputGdf, by=None):
        '''
        Constructor
        '''
        if not isinstance(inputGdf, GeoDataFrame):
            raise IllegalArgumentTypeException(inputGdf, 'GeoDataFrame')
        self.inputGdf = inputGdf
        if (by is not None) and (by not in inputGdf):
            raise Exception('%s is not a relevant field name!' % (by))
        self.by = by

    def run(self):
        geoms = []
        for geom in self.inputGdf.geometry:
            geoms += [g.buffer(0.001) for g in GeomLib.flattenGeometry(geom) if 1.0 < g.area]

        contours = []
        for geom in geoms:
            contours += GeomLib.toListOfLineStrings(geom)
        # Contour union
        contourUnion = unary_union(contours)
        # Contour network polygonization
        patches = polygonize(contourUnion)

        rows = []

        if self.by is None:
            for patch in patches:
                hits = filter(patch.representative_point().within, self.inputGdf.geometry)
                nHits = len(list(hits))
                if (0 < nHits):
                    rows.append({'geometry': patch, 'nOverlap': nHits})
        else:
            for patch in patches:
                hits = []
                for _, row in self.inputGdf.iterrows():
                    if patch.representative_point().within(row.geometry):
                        hits.append(row[self.by])
                if (0 < len(hits)):
                    rows.append({'geometry': patch, 'nOverlap': len(hits),
                                 'matched_id': ArrayCoding.encode(hits)})

        return GeoDataFrame(rows, crs=self.inputGdf.crs)



'''
Created on 3 mars 2021

@author: tleduc

Copyright 2020 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
from geopandas.geodataframe import GeoDataFrame
from t4gpd.commons.GeomLib import GeomLib
from t4gpd.commons.IllegalArgumentTypeException import IllegalArgumentTypeException
from t4gpd.commons.RayCastingLib import RayCastingLib
from t4gpd.commons.SVFLib import SVFLib
from t4gpd.morph.geoProcesses.AbstractGeoprocess import AbstractGeoprocess


class SkyViewFactorOnTopOfRoof(AbstractGeoprocess):
    '''
    classdocs
    '''

    def __init__(self, buildings, nRays=64, maxRayLen=100.0, elevationFieldname='HAUTEUR',
                 method=2018, background=True):
        '''
        Constructor
        '''
        if not isinstance(buildings, GeoDataFrame):
            raise IllegalArgumentTypeException(buildings, 'GeoDataFrame')
        self.buildings = buildings
        self.spatialIndex = buildings.sindex

        self.shootingDirs = RayCastingLib.preparePanopticRays(nRays)
        self.maxRayLen = maxRayLen

        if elevationFieldname not in buildings:
            raise Exception('%s is not a relevant field name!' % elevationFieldname)
        self.elevationFieldname = elevationFieldname

        if (1981 == method):
            print('SVF calculation method: Oke (1981), nRays = %d, maxRayLen = %.1f' % (nRays, maxRayLen))
            self.method = SVFLib.svf1981
        else:
            print('SVF calculation method: Bernard et al. (2018), nRays = %d, maxRayLen = %.1f' % (nRays, maxRayLen))
            self.method = SVFLib.svf2018

        self.background = background

    def runWithArgs(self, row):
        viewpoint = row.geometry.centroid

        enclosingFeatures = GeomLib.getEnclosingFeatures(self.buildings, self.spatialIndex, viewpoint)

        if (0 == len(enclosingFeatures)):
            # OUTDOOR/STREET-LEVEL VIEWPOINT
            _, _, hitDists, hitMasks, _ = RayCastingLib.multipleRayCast25D(
                self.buildings, self.spatialIndex, viewpoint, self.shootingDirs,
                self.maxRayLen, self.elevationFieldname, self.background)

        else:
            # INDOOR/ROOFTOP VIEWPOINT

            # REMOVE enclosingFeatures FROM THE GeoDataFrame & SpatialIndex OF SURROUNDING MASKS
            enclosingFeaturesIndex = [f.name for f in enclosingFeatures]
            _buildings = self.buildings[~self.buildings.index.isin(enclosingFeaturesIndex)]
            _buildings.reset_index(drop=True, inplace=True)  # HANDLE POTENTIAL KeyError
            _sindex = _buildings.sindex

            # ASSESS THE SVF
            _, _, hitDists, hitMasks, _ = RayCastingLib.multipleRayCastOnTopOfRoof(
                _buildings, _sindex, viewpoint, self.shootingDirs,
                self.maxRayLen, enclosingFeatures, self.elevationFieldname, self.background)

        hitHeights = [0.0 if f is None else f[self.elevationFieldname] for f in hitMasks]

        return {
            # 'hit_heights': ArrayCoding.encode(hitHeights),
            'svf_roof': self.method(hitHeights, hitDists)
            }



'''
Created on 18 juin 2020

@author: tleduc

Copyright 2020 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
from geopandas.geodataframe import GeoDataFrame
from shapely.geometry import Point

from t4gpd.commons.GeomLib import GeomLib
from t4gpd.morph.geoProcesses.AbstractGeoprocess import AbstractGeoprocess
from t4gpd.commons.IllegalArgumentTypeException import IllegalArgumentTypeException


class BiggestInscribedDisc(AbstractGeoprocess):
    '''
    classdocs
    '''

    def __init__(self, masksGdf, nsegm=-1):
        '''
        Constructor
        '''
        if not isinstance(masksGdf, GeoDataFrame):
            raise IllegalArgumentTypeException(masksGdf, 'GeoDataFrame')
        self.masksGdf = masksGdf
        self.spatialIdx = self.masksGdf.sindex
        self.nsegm = nsegm

    def runWithArgs(self, row):
        geom = row.geometry
        if not isinstance(geom, Point):
            raise IllegalArgumentTypeException(geom, 'GeoDataFrame of Point')

        buffDist = 40.0
        minDist, _, _ = GeomLib.getNearestFeature(self.masksGdf, self.spatialIdx, geom, buffDist)

        return {
            'geometry': geom.buffer(minDist, self.nsegm),
            'insc_diam': 2 * minDist
            }



'''
Created on 17 janv. 2021

@author: tleduc

Copyright 2020 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
from datetime import datetime, timezone

from geopandas.geodataframe import GeoDataFrame
from numpy import cos, sin
from shapely.affinity import translate
from shapely.geometry import LineString
from t4gpd.commons.DatetimeLib import DatetimeLib
from t4gpd.commons.GeoProcess import GeoProcess
from t4gpd.commons.IllegalArgumentTypeException import IllegalArgumentTypeException
from t4gpd.commons.sun.SunLib import SunLib


class STSunMap2D(GeoProcess):
    '''
    classdocs
    '''

    def __init__(self, viewpointsGdf, datetimes, size=4.0, projectionName='Stereographic',
                 tz=timezone.utc, model='pysolar'):
        '''
        Constructor
        '''
        if not isinstance(viewpointsGdf, GeoDataFrame):
            raise IllegalArgumentTypeException(viewpointsGdf, 'GeoDataFrame')
        self.viewpointsGdf = viewpointsGdf

        self.datetimes = DatetimeLib.generate(datetimes, tz)

        self.size = size

        if not projectionName in ['Stereographic']:
            raise IllegalArgumentTypeException(projectionName, 'spherical projection as "Stereographic"')
        self.proj = STSunMap2D.__stereographic
        self.sunModel = SunLib(gdf=viewpointsGdf, model=model)

    @staticmethod
    def __stereographic(lat, lon, size=1.0):
        radius = (size * cos(lat)) / (1.0 + sin(lat))
        return (radius * cos(lon), radius * sin(lon))

    def run(self):
        # BUILD THE SUNPATHS
        sunpaths = []
        for _lbl, _dt in self.datetimes.items():
            if isinstance(_dt, datetime):
                alti, azim = self.sunModel.getSolarAnglesInRadians(_dt, self.lat, self.lon)
                if (0 <= alti):
                    sunpaths.append({'geometry': self.proj(alti, azim, self.size),
                                     'label': str(_lbl) })

            elif isinstance(_dt, list):
                _xy = []
                for __dt in _dt:
                    alti, azim = self.sunModel.getSolarAnglesInRadians(__dt)
                    if (0 <= alti):
                        _xy.append(self.proj(alti, azim, self.size))
                if (2 <= len(_xy)):
                    sunpaths.append({'geometry': LineString(_xy),
                                     'label': str(_lbl) })

        # -----        
        rows = []
        for _, row in self.viewpointsGdf.iterrows():

            # BUILD THE FRAMEWORK
            viewPoint = row.geometry.centroid
            east = translate(viewPoint, xoff=-self.size)
            west = translate(viewPoint, xoff=self.size)
            north = translate(viewPoint, yoff=self.size)
            south = translate(viewPoint, yoff=-self.size)
            rows += [ 
                {'geometry': viewPoint.buffer(self.size).exterior, 'label': 'framework'},
                {'geometry': LineString([east, west]), 'label': 'framework'},
                {'geometry': LineString([south, north]), 'label': 'framework'},
                ]

            # ADAPT THE SUNPATHS ALREADY BUILT
            for _sunpath in sunpaths:
                _geom, _lbl = _sunpath['geometry'], _sunpath['label']
                _geom = translate(_geom, xoff=viewPoint.x, yoff=viewPoint.y)
                rows.append({'geometry': _geom, 'label': _lbl})

        return GeoDataFrame(rows, crs=self.viewpointsGdf.crs)



'''
Created on 12 mai 2021

@author: tleduc

Copyright 2020-2021 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
from numpy import log


class WindSpeedExtrapolationLib(object):
    '''
    classdocs
    '''

    @staticmethod
    def windSpeedExtrapolation(WS_ms, h=10.0):
        '''
        Ws_ms: wind speed recorded at pedestrian level (at height 1.1 m) [m.s-1]

        Calculation of wind speed in h=10 m - extrapolation of wind speed in 1.1 m after Lam et al.(2018)
        '''
        # Ws_ms_10 Wind speed in h=10 m [m.s-1]
        WS_ms_10 = WS_ms * ((log(h / 0.01)) / (log(1.1 / 0.01)))
        # TODO: WS_ms_10 = WS_ms * (log(100 * h) / log(110))
        return WS_ms_10



'''
Created on 17 sept. 2020

@author: tleduc

Copyright 2020 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
from geopandas.geodataframe import GeoDataFrame
from shapely.geometry import LineString
from t4gpd.commons.GeoProcess import GeoProcess
from t4gpd.commons.IllegalArgumentTypeException import IllegalArgumentTypeException


class STIdentifyRowOfTrees(GeoProcess):
    '''
    classdocs
    '''

    def __init__(self, treesGdf, treeIdFieldname, roadIdFieldname='road_id', roadDistFieldname='road_dist',
                 roadAbscissaFieldname='road_absc', roadSideFieldname='road_side', nTreesThreshold=4):
        '''
        Constructor
        '''
        if not isinstance(treesGdf, GeoDataFrame):
            raise IllegalArgumentTypeException(treesGdf, 'GeoDataFrame')
        self.treesGdf = treesGdf

        if treeIdFieldname not in treesGdf:
            raise Exception('%s is not a relevant field name!' % (treeIdFieldname))
        self.treeIdFieldname = treeIdFieldname

        if roadIdFieldname not in treesGdf:
            raise Exception('%s is not a relevant field name!' % (roadIdFieldname))
        self.roadIdFieldname = roadIdFieldname

        if roadDistFieldname not in treesGdf:
            raise Exception('%s is not a relevant field name!' % (roadDistFieldname))
        self.roadDistFieldname = roadDistFieldname

        if roadAbscissaFieldname not in treesGdf:
            raise Exception('%s is not a relevant field name!' % (roadAbscissaFieldname))
        self.roadAbscissaFieldname = roadAbscissaFieldname

        if roadSideFieldname not in treesGdf:
            raise Exception('%s is not a relevant field name!' % (roadSideFieldname))
        self.roadSideFieldname = roadSideFieldname

        self.nTreesThreshold = nTreesThreshold

    def run(self):
        ht = dict()

        for _, row in self.treesGdf.iterrows():
            treeGeom = row.geometry.coords[0]
            treeId = row[self.treeIdFieldname]
            roadId = row[self.roadIdFieldname]
            roadDist = row[self.roadDistFieldname]
            roadAbsc = row[self.roadAbscissaFieldname]
            roadSide = row[self.roadSideFieldname]

            currItem = {
                'tree_id': treeId,
                'road_dist': roadDist,
                'road_absc': roadAbsc,
                'geometry': treeGeom
                }

            if roadId not in ht:
                ht[roadId] = {roadSide: [currItem]}
            elif roadSide not in ht[roadId]:
                ht[roadId][roadSide] = [currItem]
            else:
                ht[roadId][roadSide].append(currItem)

        rows = []
        for roadId in ht.keys():
            for roadSide in ht[roadId].keys():
                if self.nTreesThreshold < len(ht[roadId][roadSide]):
                    ht[roadId][roadSide].sort(key=lambda t: t['road_absc'])
                    geom = LineString([t['geometry'] for t in ht[roadId][roadSide]])
                    rows.append({'geometry': geom, 'road_id': roadId, 'road_side': roadSide})

        return GeoDataFrame(rows, crs=self.treesGdf.crs)



'''
Created on 25 juil. 2022

@author: tleduc

Copyright 2020-2022 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
from t4gpd.io.AbstractReader import AbstractReader
from t4gpd.io.CirReader import CirReader
from t4gpd.io.ValReader import ValReader


class CirValReader(AbstractReader):
    '''
    classdocs
    '''

    def __init__(self, cirFilename, *valFilenames):
        '''
        Constructor
        '''
        self.cirFilename = cirFilename
        self.valFilenames = valFilenames

    def run(self):
        gdf = CirReader(self.cirFilename).run()
        for filename in self.valFilenames:
            df = ValReader(filename).run()
            gdf = gdf.merge(df, how='left', on='cir_id')
        return gdf



'''
Created on 12 nov. 2021

@author: tleduc

Copyright 2020-2021 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
import matplotlib.pyplot as plt
from numpy import inf, max, min, zeros
from t4gpd.commons.GeoProcess import GeoProcess


class PlotAMatrixOfDiagrams(GeoProcess):
    '''
    classdocs
    '''

    def __getYRanges(self, XY):
        _y0, _y1 = inf, -inf
        for k in range(1, len(XY)):
            _y0 = min([_y0, *XY[k]])
            _y1 = max([_y1, *XY[k]])
        return _y0, _y1

    def __init__(self, matrixOfXY, xlim=None, ylim=None, xLbls=None, yLbls=None,
                 suptitle=None, titles=None, bar=False, grid=True, bottom=False,
                 magn=1, outputFile=None):
        '''
        Constructor
        '''
        self.matrixOfXY = matrixOfXY
        self.ncols, self.nrows = len(self.matrixOfXY[0]), len(self.matrixOfXY)
        _COLS, _ROWS = range(self.ncols), range(self.nrows)

        _magn = 0.05
        if isinstance(xlim, (tuple, list)) and (0 == len(xlim)):
            _x0 = min([ min(self.matrixOfXY[l][c][0]) for c in _COLS for l in _ROWS])
            _x1 = max([ max(self.matrixOfXY[l][c][0]) for c in _COLS for l in _ROWS])
            _offX = _magn * (_x1 - _x0)
            xlim = [_x0 - _offX, _x1 + _offX]
        self.xlim = xlim

        if isinstance(ylim, (tuple, list)) and (0 == len(ylim)):
            _y0 = min([ self.__getYRanges(self.matrixOfXY[l][c])[0] for c in _COLS for l in _ROWS ])
            _y1 = max([ self.__getYRanges(self.matrixOfXY[l][c])[1] for c in _COLS for l in _ROWS ])
            _offY = _magn * (_y1 - _y0)
            ylim = [_y0 - _offY, _y1 + _offY]
        self.ylim = ylim

        if isinstance(xLbls, str):
            _matrix = []
            for _l in _ROWS:
                _row = []
                for _c in _COLS:
                    _row.append(xLbls if (self.nrows - 1 == _l) else None)
                _matrix.append(_row)
            xLbls = _matrix
        self.xLbls = xLbls

        if isinstance(yLbls, str):
            _matrix = []
            for _l in _ROWS:
                _row = []
                for _c in _COLS:
                    _row.append(yLbls if (0 == _c) else None)
                _matrix.append(_row)
            yLbls = _matrix
        self.yLbls = yLbls

        self.suptitle = suptitle

        if isinstance(titles, str):
            _matrix = []
            for _l in _ROWS:
                _row = []
                for _c in _COLS:
                    _row.append(titles if (0 == _l) else None)
                _matrix.append(_row)
            titles = _matrix
        self.titles = titles

        self.bar = bar
        self.grid = grid
        self.bottom = bottom
        self.magn = magn * 8.26  # 21 cm ~ 8.26 inches
        self.outputFile = outputFile

    def run(self):
        mainFig = plt.figure(figsize=(self.ncols * self.magn, self.nrows * self.magn))
        mainFig.clear()

        if not self.suptitle is None:
            mainFig.suptitle(self.suptitle)

        for fignum in range(self.nrows * self.ncols):
            nr = fignum // self.ncols
            nc = fignum % self.ncols
            plt.subplot(self.nrows, self.ncols, fignum + 1)  # numrows, numcols, fignum

            _myplot = plt.bar if self.bar else plt.plot

            _bottom = zeros(len(self.matrixOfXY[nr][nc][0]))
            for _k in range(1, len(self.matrixOfXY[nr][nc])):
                _myplot(self.matrixOfXY[nr][nc][0], self.matrixOfXY[nr][nc][_k],
                        bottom=_bottom)
                if self.bottom:
                    _bottom += self.matrixOfXY[nr][nc][_k]

            if not self.titles is None:
                plt.title(self.titles[nr][nc])
            if not self.xLbls is None:
                plt.xlabel(self.xLbls[nr][nc])
            if not self.yLbls is None:
                plt.ylabel(self.yLbls[nr][nc])
            if not self.xlim is None:
                plt.xlim(self.xlim)
            if not self.ylim is None:
                plt.ylim(self.ylim)

            plt.grid(self.grid)

        if self.outputFile is None:
            mainFig.tight_layout()
            plt.show()
        else:
            plt.savefig(self.outputFile, format='pdf', bbox_inches='tight')
        plt.close(mainFig)

'''
'''
from numpy import abs, cos, linspace, pi, sin

nr, nc = 2, 3
n = 50
X = linspace(0, 2 * pi, n)
m = [
    [(X, (l + c) * abs(sin(X)), abs(cos(X))) for c in range(nc)] for l in range(nr)
    ]
PlotAMatrixOfDiagrams(m, xlim=[], ylim=[],
                      xLbls='Absc.', yLbls='Ord.',
                      suptitle='SIN', titles='Sin',
                      bar=True, grid=True, bottom=False,
                      magn=0.25, outputFile=None).run()
'''
'''



'''
Created on 17 aug. 2020

@author: tleduc

Copyright 2020 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
from geopandas.geodataframe import GeoDataFrame
from shapely.geometry import LineString, Polygon
from t4gpd.commons.GeoProcess import GeoProcess
from t4gpd.commons.IllegalArgumentTypeException import IllegalArgumentTypeException


class ObjReader(GeoProcess):
    '''
    classdocs
    '''

    def __init__(self, inputFile, crs='EPSG:2154'):
        '''
        Constructor
        '''
        self.inputFile = inputFile
        self.crs = crs

    def __extract_v_vt_vn(self, tmp):
        # Extract vertex index/vextex texture coord. index/vertex normal index
        # f v1/vt1/vn1 v2/vt2/vn2 v3/vt3/vn3 ...
        tmp = tmp.strip().split('/')
        if 1 == len(tmp):
            return tmp[0], None, None
        elif 2 == len(tmp):
            return tmp[0], tmp[1], None
        elif 3 == len(tmp):
            return tmp
        raise IllegalArgumentTypeException(tmp, 'v1/vt1/vn1')

    def __read(self):
        rows_of_faces, rows_of_lines = [], []
        
        with open(self.inputFile, 'r') as f:
            # objectName, groupName, materialName = None, None, None
            vtxId, vertices = 1, dict()

            for line in f:
                line = line.strip()
                tmp = line.split()
                if 0 < len(tmp):
                    if 'o' == tmp[0]:
                        # objectName = tmp[1]
                        pass

                    elif 'g' == tmp[0]:
                        # groupName = tmp[1]
                        pass

                    elif 'usemtl' == tmp[0]:
                        # materialName = tmp[1]
                        pass

                    elif 'v' == tmp[0]:
                        vertices[vtxId] = (float(tmp[1]), float(tmp[2]), float(tmp[3]))
                        vtxId += 1

                    elif 'l' == tmp[0]:  # POLYLINE ELEMENTS
                        _vertices = [vertices[int(i)] for i in tmp[1:]]
                        _polyline = LineString(_vertices)
                        _row = {'geometry': _polyline}
                        rows_of_lines.append(_row)

                    elif 'f' == tmp[0]:  # POLYGONAL ELEMENTS
                        _trios = [self.__extract_v_vt_vn(x) for x in tmp[1:]]
                        _vertices = [vertices[int(i)] for i, _, _ in _trios]
                        _polygon = Polygon(_vertices)
                        _row = {'geometry': _polygon}
                        rows_of_faces.append(_row)

        return rows_of_faces, rows_of_lines

    def run(self):
        rows_of_faces, rows_of_lines = self.__read()

        if 0 == len(rows_of_faces) + len(rows_of_lines):
            return None
        elif 0 == len(rows_of_faces):
            return GeoDataFrame(rows_of_lines, crs=self.crs)
        elif 0 == len(rows_of_lines):
            return GeoDataFrame(rows_of_faces, crs=self.crs)
        return GeoDataFrame(rows_of_faces, crs=self.crs), GeoDataFrame(rows_of_lines, crs=self.crs)



'''
Created on 12 sept. 2020

@author: tleduc

Copyright 2020 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
from t4gpd.commons.GeoProcess import GeoProcess
from geopandas.geodataframe import GeoDataFrame
from t4gpd.commons.AngleLib import AngleLib


class AbstractHardShadow(GeoProcess):
    '''
    classdocs
    '''

    def run(self):
        rows = []

        for _dt, _radDir, _solarAlti, _solarAzim in self.sunPositions:
            # Hypothesis: the shadow cast by a 1m high mask is at most 10m long (arctan(0.1)=5.71) 
            if (_solarAlti < AngleLib.toRadians(5.71)):
                print('Solar height less than 5.71 degrees (%.1f degrees) on %s!' % (
                    AngleLib.toDegrees(_solarAlti), _dt))
            else:
                for _, row in self.gdf.iterrows():
                    _shadow = self._auxiliary(row, _radDir, _solarAlti, _solarAzim)
                    if _shadow is not None:
                        _row = self.updateOrAppend(row, {'datetime': str(_dt), 'geometry': _shadow})
                        rows.append(_row)

        result = GeoDataFrame(rows, crs=self.crs)

        if self.aggregate:
            result = result[['datetime', 'geometry']].dissolve(by='datetime')
            # Use a buffer to avoid slivers
            result.geometry = result.buffer(0.001, -1)
            result.reset_index(inplace=True)

        return result



'''
Created on 16 feb. 2021

@author: tleduc

Copyright 2020 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''

from geopandas.geodataframe import GeoDataFrame
from pandas import DataFrame
from t4gpd.commons.GeoProcess import GeoProcess
from t4gpd.commons.IllegalArgumentTypeException import IllegalArgumentTypeException
from windrose.windrose import WindroseAxes

import matplotlib.pyplot as plt 


class WindRose(GeoProcess):
    '''
    classdocs
    '''

    def __init__(self, inputGdf, windDirectionFieldname='WindDirection', windSpeedFieldname='WindSpeed',
                 title=None, outputFile=None, nsector=16, speedClasses=6):
        '''
        Constructor
        '''
        if not isinstance(inputGdf, (DataFrame, GeoDataFrame)):
            raise IllegalArgumentTypeException(inputGdf, 'GeoDataFrame')
        self.inputGdf = inputGdf

        for fieldname in (windSpeedFieldname, windDirectionFieldname):
            if fieldname not in inputGdf:
                raise Exception('%s is not a relevant field name!' % fieldname)
        self.windDirection = windDirectionFieldname
        self.windSpeed = windSpeedFieldname

        self.title = title
        self.outputFile = outputFile
        self.nsector = nsector
        self.speedClasses = speedClasses

    def run(self):
        ws = self.inputGdf[self.windSpeed]
        wd = self.inputGdf[self.windDirection]

        ax = WindroseAxes.from_ax()
        if self.title is not None:
            ax.set_title(self.title)
        ax.bar(wd, ws, normed=True, opening=0.8, edgecolor='white', cmap=None,
               bins=self.speedClasses, nsector=self.nsector)
        ax.set_legend(title="Wind speed (m.s$^{-1}$)")
        ax.set_xticklabels(['E', 'N-E', 'N', 'N-W', 'W', 'S-W', 'S', 'S-E'])

        if self.outputFile is None:
            plt.show()
        else:
            plt.savefig(self.outputFile, format='pdf', bbox_inches='tight')

'''
from random import randrange

rows = [{'WindDirection': randrange(0, 360), 'WindSpeed': randrange(0, 10)} for _ in range(100)]
df = DataFrame(rows)

df = DataFrame(data={
    'WindDirection': [0, 0, 0, 90],
    'WindSpeed': [8, 6, 7, 4],
    })
WindRose(df, title='XXX', nsector=16, speedClasses=3).run()
'''



'''
Created on 16 dec. 2020

@author: tleduc

Copyright 2020 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
from geopandas.geodataframe import GeoDataFrame
from shapely.geometry import LineString, MultiLineString
from shapely.geometry import Point
from t4gpd.commons.GeomLib import GeomLib
from t4gpd.commons.IllegalArgumentTypeException import IllegalArgumentTypeException
from t4gpd.commons.graph.AbstractUrbanGraphLib import AbstractUrbanGraphLib

import networkx as nx


class MCALib(object):
    '''
    classdocs
    '''

    def __init__(self, roads, roi=None):
        '''
        Constructor
        '''
        if not isinstance(roads, GeoDataFrame):
            raise IllegalArgumentTypeException(roads, 'GeoDataFrame')

        if roi is None:
            _roads = roads
            geometries = [geom for geom in roads.geometry]
        else:
            ids = roads.sindex.intersection(roi)
            geometries = [roads.loc[_id].geometry for _id in ids]
            _roads = GeoDataFrame([{'geometry': geom} for geom in geometries], crs=roads.crs)

        self.roads = _roads
        self.graph, self.icVertices = MCALib.__buildGraph(geometries)

    @staticmethod
    def __addCoupleOfVertices(graph, first, last, linestring):
        graph.add_edge(first, last, object={
            'geometry': linestring, 'weight':linestring.length})

    @staticmethod
    def __addEdge(graph, ciVertices, icVertices, linestring):
        multigraph = 0
        coords = linestring.coords
        if (2 <= len(coords)):
            first = AbstractUrbanGraphLib.addVertex(ciVertices, icVertices, coords[0])
            last = AbstractUrbanGraphLib.addVertex(ciVertices, icVertices, coords[-1])

            if (first in graph) and (last in graph[first]):
                multigraph = 1
                # Handle multigraph (two vertices may be connected by more than one edge)
                newEdges = GeomLib.cutsLineStringByCurvilinearDistance(linestring, linestring.length / 2.0)
                for _edge in newEdges:
                    MCALib.__addEdge(graph, ciVertices, icVertices, _edge)
            else:
                MCALib.__addCoupleOfVertices(graph, first, last, linestring)
                MCALib.__addCoupleOfVertices(graph, last, first, LineString(reversed(coords)))
        return multigraph

    @staticmethod
    def __buildGraph(geometries):
        graph = nx.DiGraph()
        ciVertices = dict()
        icVertices = dict()

        multigraph = 0
        for geom in geometries:
            if isinstance(geom, MultiLineString):
                for _geom in geom.geoms:
                    multigraph += MCALib.__addEdge(graph, ciVertices, icVertices, _geom)
            elif isinstance(geom, LineString):
                multigraph += MCALib.__addEdge(graph, ciVertices, icVertices, geom)
            # else:
                # raise IllegalArgumentTypeException(geom, 'LineString or MultiLineString')

        if 0 < multigraph:
            print('Handling of %d multigraph cases!' % multigraph)

        return graph, icVertices

    def __fromGraphToGeoDataFrame(self, centralities, label):
        rows = []
        for nodeIndex, centrality in centralities.items():
            rows.append({
                'geometry': Point((self.icVertices[nodeIndex])),
                label: centrality
                })
        return GeoDataFrame(rows, crs=self.roads.crs)

    def betweenness_centrality(self):
        centralities = nx.betweenness_centrality(self.graph, weight='weight')
        return self.__fromGraphToGeoDataFrame(centralities, 'betweenness')

    def closeness_centrality(self):
        centralities = nx.closeness_centrality(self.graph, distance='weight')
        return self.__fromGraphToGeoDataFrame(centralities, 'closeness')

    def degree_centrality(self):
        centralities = nx.degree_centrality(self.graph)
        return self.__fromGraphToGeoDataFrame(centralities, 'degree_c')



'''
Created on 11 juin 2020

@author: tleduc

Copyright 2020 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
from geopandas.geodataframe import GeoDataFrame
from shapely.geometry import Point, Polygon
from t4gpd.commons.GeomLib import GeomLib
from numpy import ndarray
from t4gpd.commons.IllegalArgumentTypeException import IllegalArgumentTypeException


class BoundingBox(object):
    '''
    classdocs
    '''

    def __init__(self, obj):
        '''
        Constructor
        '''
        if isinstance(obj, GeoDataFrame):
            self.minx, self.miny, self.maxx, self.maxy = obj.total_bounds
        elif GeomLib.isAShapelyGeometry(obj):
            self.minx, self.miny, self.maxx, self.maxy = obj.bounds
        elif (isinstance(obj, (ndarray, list, tuple)) and (4 == len(obj))):
            self.minx, self.miny, self.maxx, self.maxy = obj
        else:
            raise IllegalArgumentTypeException(obj, 'GeoDataFrame, a Shapely geometry or a 4-uple')
        self.__center = None

    def center(self):
        if self.__center is None:
            cx, cy = (self.minx + self.maxx) / 2.0, (self.miny + self.maxy) / 2.0
            self.__center = Point(cx, cy) 
        return self.__center

    def height(self):
        return (self.maxy - self.miny)

    def width(self):
        return (self.maxx - self.minx)
    
    def asPolygon(self):
        return Polygon([
            (self.minx, self.miny),
            (self.minx, self.maxy),
            (self.maxx, self.maxy),
            (self.maxx, self.miny)
            ])

    def __str__(self):
        return '%s:: (%f, %f) -> (%f, %f)' % (
            __name__,
            self.minx, self.miny,
            self.maxx, self.maxy
            )



'''
Created on 29 sept. 2020

@author: tleduc

Copyright 2020 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
from pandas import DataFrame
from t4gpd.commons.GeoProcess import GeoProcess
from t4gpd.commons.IllegalArgumentTypeException import IllegalArgumentTypeException

import matplotlib.pyplot as plt


class FrequencyHistogram(GeoProcess):
    '''
    classdocs
    '''

    def __init__(self, inputGdf, fieldnames, nClusters='auto', rangeValues=None, outputFile='',
                 density=False, cumulative=False, title=None, fontsize=20):
        '''
        Constructor
        '''
        if not isinstance(inputGdf, DataFrame):
            raise IllegalArgumentTypeException(inputGdf, 'DataFrame')
        self.inputGdf = inputGdf

        if isinstance(fieldnames, str):
            self.fieldnames = [fieldnames]
        elif isinstance(fieldnames, (list, tuple)):
            self.fieldnames = fieldnames
        else:
            raise IllegalArgumentTypeException(fieldnames, 'must be a list or tuple of fieldnames')
        for fieldname in self.fieldnames:
            if fieldname not in self.inputGdf:
                raise Exception('%s is not a relevant field name!' % (fieldname))
        self.nFields = len(self.fieldnames)

        self.nClusters = nClusters if isinstance(nClusters, (list, tuple)) else [nClusters] * self.nFields

        if rangeValues is None:
            self.rangeValues = [None] * self.nFields
        elif isinstance(rangeValues, (list, tuple)):
            if ((2 == len(rangeValues)) and 
                isinstance(rangeValues[0], (int, float)) and
                isinstance(rangeValues[1], (int, float))):
                self.rangeValues = [rangeValues] * self.nFields
            else:
                self.rangeValues = rangeValues

        if (self.nFields != len(self.nClusters)):
            raise IllegalArgumentTypeException(
                self.nClusters, 'nClusters must be either an int or a list of %d items' % self.nFields)
        if (self.nFields != len(self.rangeValues)):
            raise IllegalArgumentTypeException(
                self.rangeValues, 'rangeValues must be either a float or a list of %d items' % self.nFields)

        self.outputFile = outputFile
        self.density = density
        self.cumulative = cumulative
        self.title = title
        self.fontsize = fontsize

    def run(self):
        values = [list(self.inputGdf[fieldname]) for fieldname in self.fieldnames]

        numcols, numrows, pairsOfWeightsAndRanges = 1, self.nFields, []

        mainFig = plt.figure(figsize=(numcols * 8.26, numrows * 8.26))  # 21 cm ~ 8.26 inches
        mainFig.clear()

        for fignum in range(self.nFields):
            plt.subplot(numrows, numcols, fignum + 1)  # numrows, numcols, fignum     
            plt.xlabel(self.fieldnames[fignum], fontsize=self.fontsize)
            if self.density:
                plt.ylabel('Probability density', fontsize=self.fontsize)
            else:
                plt.ylabel('Number of values', fontsize=self.fontsize)
            plt.grid(True)

            n, bins, _ = plt.hist(values[fignum], bins=self.nClusters[fignum],
                                  density=self.density, cumulative=self.cumulative,
                                  range=self.rangeValues[fignum],
                                  facecolor='gray', edgecolor='black', orientation='vertical',
                                  hatch='', histtype='bar', align='mid', rwidth=1.0)
            x = [(bins[i] + bins[i + 1]) / 2.0 for i in range(len(bins) - 1)]
            pairsOfWeightsAndRanges.append([n, bins])
            if self.title is None:
                plt.title('Histogram of %s (%d bins)' % (self.fieldnames[fignum], len(n)),
                          fontsize=self.fontsize + 4)
            else:
                plt.title(self.title, fontsize=self.fontsize + 4)
            # plt.plot(x, n, 'ro', x, n, 'r--')

        if self.outputFile is not None:
            if ('' == self.outputFile):
                plt.show()
            else:
                plt.savefig(self.outputFile, format='pdf', bbox_inches='tight')  # , facecolor='w', edgecolor='k', pad_inches=0.15)

        plt.close(mainFig)
        return pairsOfWeightsAndRanges



'''
Created on 28 sept. 2020

@author: tleduc

Copyright 2020 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
from geopandas.geodataframe import GeoDataFrame
from numpy import mean
from t4gpd.commons.IllegalArgumentTypeException import IllegalArgumentTypeException
from t4gpd.commons.RayCastingLib import RayCastingLib
from t4gpd.morph.geoProcesses.AbstractGeoprocess import AbstractGeoprocess


class WMean(AbstractGeoprocess):
    '''
    classdocs
    '''

    def __init__(self, buildingsGdf, nRays=64, maxRayLen=100.0):
        '''
        Constructor
        '''
        if not isinstance(buildingsGdf, GeoDataFrame):
            raise IllegalArgumentTypeException(buildingsGdf, 'GeoDataFrame')
        self.buildingsGdf = buildingsGdf
        self.spatialIndex = buildingsGdf.sindex

        self.shootingDirs = RayCastingLib.preparePanopticRays(nRays)
        self.maxRayLen = maxRayLen

    def runWithArgs(self, row):
        viewPoint = row.geometry.centroid

        _, _, hitDists, _ = RayCastingLib.multipleRayCast2D(
            self.buildingsGdf, self.spatialIndex, viewPoint, self.shootingDirs, self.maxRayLen)

        return {
            # 'hit_dists': ArrayCoding.encode(hitDists),
            'wmean': float(mean(hitDists))
            }



'''
Created on 31 dec. 2020

@author: tleduc

Copyright 2020 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
from geopandas.geodataframe import GeoDataFrame
from shapely.ops import polygonize, unary_union
from t4gpd.commons.GeoProcess import GeoProcess
from t4gpd.commons.GeomLib import GeomLib
from t4gpd.commons.IllegalArgumentTypeException import IllegalArgumentTypeException


class STPolygonize(GeoProcess):
    '''
    classdocs
    '''

    def __init__(self, inputGdf):
        '''
        Constructor
        '''
        if not isinstance(inputGdf, GeoDataFrame):
            raise IllegalArgumentTypeException(inputGdf, 'GeoDataFrame')
        self.inputGdf = inputGdf
        
    def run(self):
        contours = []
        for geom in self.inputGdf.geometry:
            contours += GeomLib.toListOfLineStrings(geom)

        # Contour union
        contourUnion = unary_union(contours)
        # Contour network polygonization
        patches = polygonize(contourUnion)

        rows = [{'gid': i, 'geometry': patch} for i, patch in enumerate(patches)]
        return GeoDataFrame(rows, crs=self.inputGdf.crs)




'''
Created on 11 juin 2020

@author: tleduc

Copyright 2020 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
from geopandas.geodataframe import GeoDataFrame
from t4gpd.commons.GeoProcess import GeoProcess
from t4gpd.commons.IllegalArgumentTypeException import IllegalArgumentTypeException
from t4gpd.commons.grid.GridLib import GridLib


class STGrid(GeoProcess):
    '''
    classdocs
    '''

    def __init__(self, inputGdf, dx, dy=None, indoor=None, intoPoint=True, encode=True):
        '''
        Constructor
        '''
        if not isinstance(inputGdf, GeoDataFrame):
            raise IllegalArgumentTypeException(inputGdf, 'GeoDataFrame')

        self.inputGdf = inputGdf
        self.dx = dx
        self.dy = dy
        if indoor in [None, True, False, 'both']:
            self.indoor = indoor
        else:
            raise Exception('Illegal argument: indoor must be chosen in [None, True, False, "both"]!')
        self.intoPoint = intoPoint
        self.encode = encode

    def run(self):
        gridLib = GridLib(self.inputGdf, self.dx, self.dy, self.encode)

        if self.indoor is None:
            grid = gridLib.grid()
        elif ('both' == self.indoor):
            grid = gridLib.indoorOutdoorGrid()
        elif self.indoor:
            grid = gridLib.indoorGrid()
        else:
            grid = gridLib.outdoorGrid()

        if self.intoPoint:
            grid.geometry = grid.centroid
        return grid



'''
Created on 31 aug. 2020

@author: tleduc

Copyright 2020 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
from geopandas.geodataframe import GeoDataFrame
from shapely.wkt import loads
from t4gpd.commons.CSVLib import CSVLib
from t4gpd.commons.GeoProcess import GeoProcess


class CSVWKTReader(GeoProcess):
    '''
    classdocs
    '''

    def __init__(self, inputFile, geomFieldName='the_geom:GEOMETRY', fieldSep=';',
                 srcEpsgCode='EPSG:4326', dstEpsgCode=None, decimalSep='.'):
        '''
        Constructor
        '''
        self.inputFile = inputFile
        self.geomFieldName = geomFieldName
        self.fieldSep = fieldSep
        self.crs = srcEpsgCode
        self.dstEpsgCode = dstEpsgCode
        self.decimalSep = decimalSep
        
    def run(self):
        _rows = CSVLib.read(self.inputFile, self.fieldSep, self.decimalSep)

        rows = []
        for row in _rows:
            row['geometry'] = loads(row[self.geomFieldName])
            rows.append(row)

        outputGdf = GeoDataFrame(rows, crs=self.crs)
        if 'geometry' != self.geomFieldName:
            outputGdf = outputGdf.drop(self.geomFieldName, axis=1)

        if not self.dstEpsgCode is None:
            outputGdf = outputGdf.to_crs(self.dstEpsgCode)
        return outputGdf



'''
Created on 22 oct. 2020

@author: tleduc

Copyright 2020 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
import re
from shapely.geometry import Polygon

from geopandas.geodataframe import GeoDataFrame
from t4gpd.commons.CSVLib import CSVLib
from t4gpd.commons.GeoProcess import GeoProcess


class CityGMLReader(GeoProcess):
    '''
    classdocs
    '''
    RE_BUILDING = re.compile(r'<bldg:Building\s')
    RE_OUTER_BUILDING_INSTALLATION = re.compile(r'<bldg:outerBuildingInstallation')
    RE_GROUND_SURFACE = re.compile(r'<bldg:GroundSurface')
    RE_ROOF_SURFACE = re.compile(r'<bldg:RoofSurface')
    RE_WALL_SURFACE = re.compile(r'<bldg:WallSurface')
    RE_DOOR = re.compile(r'<bldg:Door')
    RE_WINDOW = re.compile(r'<bldg:Window')
    RE_RELIEF_FEATURE = re.compile(r'<dem:ReliefFeature')
    RE_TIN_RELIEF = re.compile(r'<dem:TINRelief')
    RE_POLYGON = re.compile(r'<gml:Polygon')
    RE_EXTERIOR_RING = re.compile(r'<gml:exterior')
    RE_INTERIOR_RING = re.compile(r'<gml:interior')
    RE_POS_LIST = re.compile(r'<gml:posList')
    RE_EXTRACT_NODES = re.compile(r'<gml:posList[^>]*>(.*)<\/gml:posList>')
    RE_EXTRACT_POLYGON_ID = re.compile(r'<gml:Polygon[^>]*gml:id="([^"]*)"[^>]*>')
  
    def __init__(self, inputFile, srcEpsgCode='EPSG:4326', dstEpsgCode=None):
        '''
        Constructor
        '''
        self.inputFile = inputFile
        self.crs = srcEpsgCode
        self.dstEpsgCode = dstEpsgCode

    def __fromPosListToArrayOfPoint3d(self, txt):
        coords = txt.strip().split()
        return [
            (CSVLib.readLexeme(coords[i]),
             CSVLib.readLexeme(coords[i + 1]),
             CSVLib.readLexeme(coords[i + 2])) for i in range(0, len(coords), 3)
            ]

    def __version2(self):
        the_type, subType, subSubType = None, None, None
        polygonId, objectId, subObjectId = 1, 1, 1
        extNodes, holesNodes = None, None
        isExteriorRing = None

        rows = []
        with open(self.inputFile, 'r') as f:
            for line in f:
                line = line.strip()

                if self.RE_BUILDING.match(line):
                    the_type = 'Building_%d' % (objectId)
                    subType, subSubType = None, None
                    objectId += 1
                elif self.RE_OUTER_BUILDING_INSTALLATION.match(line):
                    # TODO
                    pass
                elif self.RE_GROUND_SURFACE.match(line):
                    subType = 'GroundSurface_%d' % (subObjectId)
                    subSubType = None
                    subObjectId += 1
                elif self.RE_ROOF_SURFACE.match(line):
                    subType = 'RoofSurface_%d' % (subObjectId)
                    subSubType = None
                    subObjectId += 1
                elif self.RE_WALL_SURFACE.match(line):
                    subType = 'WallSurface_%d' % (subObjectId)
                    subSubType = None
                    subObjectId += 1
                elif self.RE_DOOR.match(line):
                    subSubType = 'Door'
                elif self.RE_WINDOW.match(line):
                    subSubType = 'Window'
                elif self.RE_RELIEF_FEATURE.match(line):
                    the_type = 'ReliefFeature'
                    subType, subSubType = None, None
                elif self.RE_TIN_RELIEF.match(line):
                    subType = 'TINRelief'
                elif self.RE_POLYGON.match(line):
                    tmp = self.RE_EXTRACT_POLYGON_ID.search(line.strip()).group(1)
                    polygonId = None if tmp is None else tmp[1]
                    if extNodes is not None:
                        rows.append({
                            'type': the_type,
                            'subType': subType,
                            'subSubType': subSubType,
                            'polygonId': polygonId,
                            'geometry': Polygon(extNodes, holesNodes)
                            })
                        extNodes, holesNodes = None, None

                elif self.RE_EXTERIOR_RING.match(line):
                    isExteriorRing = True
                elif self.RE_INTERIOR_RING.match(line):
                    isExteriorRing = False

                elif self.RE_POS_LIST.match(line):
                    tmp = self.RE_EXTRACT_NODES.search(line.strip()).group(1)
                    if tmp is not None:
                        if isExteriorRing:
                            extNodes = self.__fromPosListToArrayOfPoint3d(tmp)
                        else:
                            if holesNodes is None:
                                holesNodes = []
                            holesNodes.append(self.__fromPosListToArrayOfPoint3d(tmp))
        return rows

    def run(self):
        rows = self.__version2()
        outputGdf = GeoDataFrame(rows, crs=self.crs)
        if not self.dstEpsgCode is None:
            outputGdf = outputGdf.to_crs(self.dstEpsgCode)
        return outputGdf



'''
Created on 15 janv. 2021

@author: tleduc

Copyright 2020 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
from numpy import mean, std
from shapely.wkt import loads
from t4gpd.commons.GeomLib import GeomLib
from t4gpd.morph.geoProcesses.AbstractGeoprocess import AbstractGeoprocess

from t4gpd.commons.Entropy import Entropy


class StarShapedIndices(AbstractGeoprocess):
    '''
    classdocs
    '''

    def __init__(self, precision=1.0, base=2):
        '''
        Constructor
        '''
        self.precision = precision
        self.base = base

    def runWithArgs(self, row):
        geom = row.geometry

        lengths = GeomLib.fromMultiLineStringToLengths(geom)

        drift = loads(row['vect_drift']).length if ('vect_drift' in row) else None
        h = Entropy.createFromDoubleValuesArray(lengths, self.precision).h(self.base)

        return {
            'min_raylen': min(lengths),
            'avg_raylen': mean(lengths),
            'std_raylen': std(lengths),
            'max_raylen': max(lengths),
            'entropy': h,
            'drift': drift
            }



'''
Created on 3 juil. 2020

@author: tleduc

Copyright 2020 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
from numpy import cos, sin

from geopandas.geodataframe import GeoDataFrame
from shapely.geometry import Point
from t4gpd.commons.AngleLib import AngleLib
from t4gpd.commons.CSVLib import CSVLib
from t4gpd.commons.GeoProcess import GeoProcess


class CSVInertialSensorReader(GeoProcess):
    '''
    classdocs
    '''

    def __init__(self, inputFile, xFieldName='X', yFieldName='Y', zFieldName='Z',
                 distFieldName='Distance', azimFieldName='degree',
                 lon0FieldName='longitude', lat0FieldName='latitude',
                 fieldSep=',', srcEpsgCode='EPSG:4326', dstEpsgCode='EPSG:2154',
                 cartesian=True, decimalSep='.'):
        '''
        Constructor
        '''
        self.inputFile = inputFile

        if ((xFieldName is None) or (yFieldName is None) or
            (distFieldName is None) or (azimFieldName is None)):
            raise Exception('')

        self.xFieldName = xFieldName
        self.yFieldName = yFieldName
        self.zFieldName = zFieldName
        self.distFieldName = distFieldName
        self.azimFieldName = azimFieldName

        self.cartesian = cartesian
        self.lon0FieldName = lon0FieldName
        self.lat0FieldName = lat0FieldName
        self.fieldSep = fieldSep
        self.crs = srcEpsgCode
        self.dstEpsgCode = dstEpsgCode
        self.decimalSep = decimalSep

    @staticmethod
    def __transform(pt, srcEpsgCode, dstEpsgCode):
        '''
        import pyproj
        from shapely.ops import transform
        srcCrs = pyproj.CRS(srcEpsgCode)
        dstCrs = pyproj.CRS(dstEpsgCode)
        project = pyproj.Transformer.from_crs(srcCrs, dstCrs, always_xy=True).transform
        return transform(project, pt)
        '''
        inputGdf = GeoDataFrame([{'geometry':pt}], crs=srcEpsgCode)
        outputGdf = inputGdf.to_crs(dstEpsgCode)
        return outputGdf.geometry[0]

    def run(self):
        _rows = CSVLib.read(self.inputFile, self.fieldSep, self.decimalSep)

        p0 = Point(0, 0)
        if ((self.lon0FieldName is not None) and (self.lat0FieldName is not None)):
            p0 = Point(_rows[0][self.lon0FieldName], _rows[0][self.lat0FieldName])
            p0 = CSVInertialSensorReader.__transform(p0, self.crs, self.dstEpsgCode)

        rows = []
        if self.cartesian:
            for row in _rows:
                if self.zFieldName is None:
                    row['geometry'] = Point(row[self.yFieldName] + p0.x,
                                            row[self.xFieldName] + p0.y)
                else:
                    row['geometry'] = Point(row[self.yFieldName] + p0.x,
                                            row[self.xFieldName] + p0.y,
                                            row[self.zFieldName])
                rows.append(row)
                result = GeoDataFrame(rows, crs=self.dstEpsgCode)
                # result = result.rotate(_rows[0]['degree'], origin=p0, use_radians=False)

        else:
            for i, row in enumerate(_rows):
                r = row[self.distFieldName]
                azim = AngleLib.toRadians(row[self.azimFieldName])
                if (0 < i):
                    r = r - _rows[i - 1][self.distFieldName]
                row['geometry'] = p0 = Point(r * sin(azim) + p0.x, r * cos(azim) + p0.y)
                rows.append(row)
                result = GeoDataFrame(rows, crs=self.dstEpsgCode)

        return result



'''
Created on 27 mai 2021

@author: tleduc

Copyright 2020-2021 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
from datetime import date, datetime, timezone

from t4gpd.commons.IllegalArgumentTypeException import IllegalArgumentTypeException


class DateRangeLib(object):
    '''
    classdocs
    '''

    @staticmethod
    def preprocess(date0, date1=None, period='hourly', tz=timezone.utc):
        if not isinstance(date0, date):
            raise IllegalArgumentTypeException(date0, 'date')
        date0 = datetime(*date0.timetuple()[:3], 0, tzinfo=tz)

        if not (date1 is None or isinstance(date1, date)):
            raise IllegalArgumentTypeException(date1, 'date')
        if date1 is None:
            date1 = datetime(*date0.timetuple()[:3], 23, tzinfo=tz)
        else:
            date1 = datetime(*date1.timetuple()[:3], 23, tzinfo=tz)

        if (date0 > date1):
            raise Exception(f'Nonsense: date0={date0} > date1={date1}')

        if not period in ('hourly', 'daily', 'weekly', 'monthly'):
            raise Exception("Illegal argument: period must be chosen in {'hourly', 'daily', 'weekly', 'monthly'}!")

        if 'hourly' == period:
            keyFilter = lambda dt: ((dt - date0).total_seconds() / 3600)
            fieldname = 'hour'
        elif 'daily' == period:
            keyFilter = lambda dt: dt.timetuple().tm_yday
            fieldname = 'day'
        elif 'weekly' == period:
            keyFilter = lambda dt: dt.isocalendar()[1]
            fieldname = 'week'
        elif 'monthly' == period:
            keyFilter = lambda dt: dt.month
            fieldname = 'month'

        return date0, date1, keyFilter, fieldname



'''
Created on 21 juil. 2022

@author: tleduc

Copyright 2020-2022 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
from geopandas import GeoDataFrame
from numpy import asarray
from shapely.geometry import MultiLineString, MultiPoint, Point
from t4gpd.commons.ArrayCoding import ArrayCoding
from t4gpd.commons.IllegalArgumentTypeException import IllegalArgumentTypeException
from t4gpd.morph.geoProcesses.AbstractGeoprocess import AbstractGeoprocess
from t4gpd.pyvista.ToUnstructuredGrid import ToUnstructuredGrid
from t4gpd.pyvista.commons.RayCasting3DLib import RayCasting3DLib
from t4gpd.pyvista.commons.RayCastingIn3DLib import RayCastingIn3DLib


class OrientedRayCasting3D(AbstractGeoprocess):
    '''
    classdocs
    '''

    def __init__(self, listOfMasks, maskPkFieldname=None,
                 viewpoints=None, mainDirectionFieldname=None, openness=90,
                 nrays=1000, maxRayLen=None, showHitPoints=False, encode=True):
        '''
        Constructor
        '''
        assert isinstance(listOfMasks, (list, tuple)), 'listOfMasks must be a list of GeoDataFrames'
        assert all(isinstance(gdf, GeoDataFrame) for gdf in listOfMasks), 'listOfMasks must be a list of GeoDataFrames'

        if maskPkFieldname is not None:
            assert all((maskPkFieldname in gdf) for gdf in listOfMasks), f'{maskPkFieldname} must be a GeoDataFrame field name!'
        self.maskPkFieldname = maskPkFieldname

        assert isinstance(viewpoints, GeoDataFrame), 'viewpoints must be a GeoDataFrame'
        assert (mainDirectionFieldname in viewpoints), 'mainDirectionFieldname must be a viewpoints field name'
        self.mainDirectionFieldname = mainDirectionFieldname

        masks = ToUnstructuredGrid(listOfMasks, fieldname=maskPkFieldname).run()
        # self.bounds = masks.bounds

        self.obbTree = RayCastingIn3DLib.prepareVtkOBBTree(masks)

        self.openness = openness
        self.nrays = nrays
        self.maxRayLen = masks.length if maxRayLen is None else maxRayLen

        self.showHitPoints = showHitPoints
        self.encode = encode

    '''
    @staticmethod
    def __clipWithBounds(rays, bounds):
        _rays = GeoDataFrame(data=[{'geometry': MultiLineString(rays)}])
        _rays = _rays.explode(ignore_index=True)
        _rays = ToUnstructuredGrid([_rays]).run()
        _rays = _rays.clip_box(*bounds)
        _rays = [_rays.cell_points(i) for i in range(_rays.n_cells)]
        return _rays
    '''

    def runWithArgs(self, row):
        geom = row.geometry
        if not isinstance(geom, Point):
            raise IllegalArgumentTypeException(geom, 'GeoDataFrame of Point')

        main_direction = row[self.mainDirectionFieldname]

        shootingDirs = RayCasting3DLib.prepareOrientedRandomRays(
            self.nrays, main_direction, self.openness)
        shootingDirs *= self.maxRayLen 

        srcPt = asarray(geom.coords[0])
        dstPts = srcPt + shootingDirs
        srcPts = srcPt.reshape(1, -1).repeat(len(dstPts), axis=0)

        rays, hitPoints, hitDists, hitGids = RayCastingIn3DLib.mraycastObbTree(
            self.obbTree, srcPts, dstPts, self.maskPkFieldname)

        # rays = OrientedRayCasting3D.__clipWithBounds(rays, self.bounds)

        if self.maskPkFieldname is None:
            return {
                'geometry': MultiPoint(hitPoints) if (self.showHitPoints) else MultiLineString(rays),
                'n_rays': len(shootingDirs),
                'hitDists': ArrayCoding.encode(hitDists) if self.encode else hitDists,
                'inf_ratio': len([d for d in hitDists if (d == RayCastingIn3DLib.INFINITY)]) / len(hitDists)
                }

        return {
            'geometry': MultiPoint(hitPoints) if (self.showHitPoints) else MultiLineString(rays),
            'n_rays': len(shootingDirs),
            'hitDists': ArrayCoding.encode(hitDists) if self.encode else hitDists,
            'hitGids': ArrayCoding.encode(hitGids) if self.encode else hitGids,
            'inf_ratio': len([d for d in hitDists if (d == RayCastingIn3DLib.INFINITY)]) / len(hitDists)
            }



'''
Created on 12 feb. 2021

@author: tleduc

Copyright 2020 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
from geopandas import GeoDataFrame, overlay
from pandas import concat
from shapely.geometry import MultiPolygon, Polygon
from shapely.ops import unary_union
from t4gpd.commons.GeoProcess import GeoProcess
from t4gpd.commons.IllegalArgumentTypeException import IllegalArgumentTypeException
from t4gpd.morph.STPolygonize import STPolygonize


class STMakeBlocks(GeoProcess):
    '''
    classdocs
    '''

    def __init__(self, buildings, roads, roi=None):
        '''
        Constructor
        '''
        if not isinstance(buildings, GeoDataFrame):
            raise IllegalArgumentTypeException(buildings, 'GeoDataFrame')
        self.buildings = buildings

        if not isinstance(roads, GeoDataFrame):
            raise IllegalArgumentTypeException(roads, 'GeoDataFrame')
        self.roads = roads

        if not((roi is None) or isinstance(roi, GeoDataFrame)):
            raise IllegalArgumentTypeException(roi, 'GeoDataFrame or None')
        self.roi = roi

    @staticmethod
    def __mkBlocks(buildings, roads, roi):
        # POLYGONIZE ROADS
        if not roi is None:
            _roads = GeoDataFrame(concat([roads[['geometry']], roi[['geometry']]]), crs=roads.crs)
        else:
            _roads = roads
        _polygons = STPolygonize(_roads).run()

        # GROUP BUILDINGS PER URBAN BLOCKS
        _buildings = overlay(buildings[['geometry']], _polygons, how='intersection')
        _blocks = _buildings.dissolve(by='gid', as_index=False)

        # TRY TO REMOVE CONCAVITIES IN EACH BLOCK
        roadsIdx = roads.sindex
        _blocks.geometry = _blocks.geometry.apply(lambda g: STMakeBlocks.__deleteConcavities(g, roads, roadsIdx))

        return _blocks

    @staticmethod
    def __deleteConcavities(block, roads, roadsIdx):
        _block = block.convex_hull
        _subsetOfRoads = roads.loc[roadsIdx.intersection(_block.bounds)]

        if (0 == len(list(filter(_block.intersects, _subsetOfRoads.geometry)))):
            return _block

        _concavities = _block.difference(block)
        if isinstance(_concavities, Polygon):
            _concavities = MultiPolygon([_concavities])
        result = [block]
        for _concavity in _concavities.geoms:
            if (0 == len(list(filter(_concavity.intersects, _subsetOfRoads.geometry)))):
                result.append(_concavity)
        # return unary_union(result)
        result = unary_union(result)
        '''
        for buffDist in [60, 50, 40, 30, 20, 10]:
            _tmp = result.buffer(buffDist, join_style=JOIN_STYLE.round).buffer(-buffDist, join_style=JOIN_STYLE.bevel)
            if (0 == len(list(filter(_tmp.intersects, _subsetOfRoads.geometry)))):
                result = _tmp
                print(buffDist)
                break
        '''
        return result

    def run(self):
        return STMakeBlocks.__mkBlocks(self.buildings, self.roads, self.roi)

'''
rootdir = '/home/tleduc/Dropbox/crenau/2_papiers_a_ecrire/2021_wavelet/dev/data-p9'
roads = gpd.read_file(rootdir + '/roads.shp')
roi = gpd.read_file(rootdir + '/roi.shp')
buildings = gpd.read_file(rootdir + '/buildings.shp')

# buildings = GeoDataFrameDemos.singleBuildingInNantes()
# print(buildings.geometry.squeeze().wkt)
# roads = gpd.GeoDataFrame([{'geometry': loads('LINESTRING (353880 6695040, 353905 6695040)')}], crs=buildings.crs)
# roads = gpd.GeoDataFrame([{'geometry': loads('LINESTRING (353935 6695030, 353949 6695030)')}], crs=buildings.crs)
# roads = gpd.GeoDataFrame([{'geometry': loads('LINESTRING (353942 6695020, 353950 6695015)')}], crs=buildings.crs)
# roads.to_file('/home/tleduc/Dropbox/crenau/2_papiers_a_ecrire/2021_wavelet/dev/data-p9/roads3.shp')

blocks = STMakeBlocks(buildings, roads, roi).run()
blocks.to_file(rootdir + '/_blocks2.shp')

print('THE END')
'''



'''
Created on 23 juin 2020

@author: tleduc

Copyright 2020 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
from numpy import isnan, pi

from t4gpd.commons.AngleLib import AngleLib
from t4gpd.commons.ellipse.EllipseLib import EllipseLib
from t4gpd.commons.ellipse.EllipticHullLib import EllipticHullLib
from t4gpd.morph.geoProcesses.AbstractGeoprocess import AbstractGeoprocess


class MABE(AbstractGeoprocess):
    '''
    classdocs
    '''

    def __init__(self, npoints=40, threshold=None, debug=False):
        '''
        Constructor
        '''
        self.npoints = npoints
        self.mabe = EllipticHullLib(threshold, debug)

    def runWithArgs(self, row):
        geom = row.geometry
        if geom is None:
            return None

        centre, semiXAxis, semiYAxis, azim, _, _ = self.mabe.getMinimumAreaBoundingEllipse(geom)
        mabe = EllipseLib.getEllipseContour(centre, semiXAxis, semiYAxis, azim, self.npoints)

        if mabe is None:
            raise Exception('Minimum-area bounding ellipse is None!')

        mabeArea = pi * semiXAxis * semiYAxis
        if (isnan(mabeArea)) or (mabeArea < geom.area):
            raise Exception('Minimum-area bounding ellipse is %s (input geom area: %.1f)!' % (str(mabeArea), geom.area()))
        mabePerim = EllipseLib.getEllipsePerimeter(semiXAxis, semiYAxis)

        axis = EllipseLib.getEllipseMainAxis(centre, semiXAxis, semiYAxis, azim)

        return {
            'mabe_centre': str(centre),
            'mabe_axis': str(axis),
            'mabe_area': mabeArea,
            'mabe_perim': mabePerim,
            'mabe_xaxis': semiXAxis,
            'mabe_yaxis': semiYAxis,
            'mabe_azim': AngleLib.toDegrees(azim),
            'geometry': mabe
            }



'''
Created on 23 oct. 2020

@author: tleduc

Copyright 2020 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
from random import randrange
from shapely.affinity import rotate

from geopandas.geodataframe import GeoDataFrame
from numpy import arctan2, cos, pi, sin
from shapely.geometry import Point, Polygon
from shapely.ops import unary_union


class LandoltRingLib(object):
    '''
    classdocs
    '''

    @staticmethod
    def c_optotype(centre, diameter, angleInDegrees=0, npoints=40):
        r = diameter / 2
        angle0 = arctan2(diameter / 10, r)
        deltaAngle = 2 * (pi - angle0) / (npoints - 1)
        ctrPoints = [
            (centre.x + r * cos(angle0 + i * deltaAngle), centre.y + r * sin(angle0 + i * deltaAngle))
            for i in range(npoints)
            ]

        r = 3 * diameter / 10
        angle0 = arctan2(diameter / 10, r)
        deltaAngle = 2 * (pi - angle0) / (npoints - 1)
        ctrPoints += [
            (centre.x + r * cos(angle0 + i * deltaAngle), centre.y + r * sin(angle0 + i * deltaAngle))
            for i in range(npoints)
            ][::-1]

        optotype = Polygon(ctrPoints)
        if 0 == angleInDegrees:
            return optotype
        return rotate(optotype, angleInDegrees, origin=centre)

    @staticmethod
    def c_chart(diameter, nrows=2, ncols=5, npoints=40):
        x0, y0, rows = diameter, diameter, []

        for l in range(nrows):
            for c in range(ncols):
                angleInDegrees = randrange(0, 360, 45)
                centre = Point((x0 + c * 2 * diameter, y0 + l * 2 * diameter))
                geom = LandoltRingLib.c_optotype(centre, diameter, angleInDegrees, npoints=npoints)
                rows.append({'gid': l * ncols + c, 'row':l, 'col':c, 'geometry':geom})

        return GeoDataFrame(rows)

    @staticmethod
    def a4landscape_c_chart(negativeContrast=True, diameter=None, nrows=2, ncols=5, npoints=40):
        width, height = 842, 595  # A4 landscape
        if diameter is None:
            diameter = min(height / nrows, width / ncols) / 2 
        x0, y0 = diameter, diameter

        holes = []
        for l in range(nrows):
            for c in range(ncols):
                angleInDegrees = randrange(0, 360, 45)
                centre = Point((x0 + c * 2 * diameter, y0 + l * 2 * diameter))
                holes.append(LandoltRingLib.c_optotype(centre, diameter, angleInDegrees, npoints=npoints))

        bbCoords = [(0, 0), (width, 0), (width, height), (0, height)]
        if negativeContrast:
            geom = Polygon(bbCoords)
            rows = [{ 'geometry':  geom.difference(unary_union(holes)) }]
        else:
            geom = Polygon(bbCoords, [[(1, 1), (1, height - 1), (width - 1, height - 1), (width - 1, 1)]])
            rows = [{ 'geometry':  geom.union(unary_union(holes)) }]
        return GeoDataFrame(rows)

'''
from t4gpd.io.SVGWriter import SVGWriter
gdf = LandoltRingLib.a4landscape_c_chart(True)
SVGWriter(gdf, './n_c_chart.svg').run()
# inkscape --export-pdf=n_c_chart.pdf n_c_chart.svg

gdf = LandoltRingLib.a4landscape_c_chart(False)
SVGWriter(gdf, './p_c_chart.svg').run()
# inkscape --export-pdf=p_c_chart.pdf p_c_chart.svg
'''



'''
Created on 12 mai 2021

@author: tleduc

Copyright 2020-2021 Thomas Leduc

This file is part of t4gpd.

t4gpd is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t4gpd is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t4gpd.  If not, see <https://www.gnu.org/licenses/>.
'''
from t4gpd.comfort.algo.ConstantsLib import ConstantsLib


class TmrtGlobeTemperatureLib(object):
    '''
    classdocs
    '''

    @staticmethod
    def assess_tmrt(AirTC, GlobeTC, WS_ms, GlobeEmissivity=ConstantsLib.epsi_glo,
                   GlobeDiameter=ConstantsLib.D_glo):
        # calculation of the mean radiant temperature [C]
        # Mean radiant temperature after (Thorsson et al., 2007) based on observed globe temperature:
        T_mrt = ((GlobeTC + ConstantsLib.T_kel) ** 4 + 
                 ((1.1e8 * WS_ms ** 0.6) / (GlobeEmissivity * GlobeDiameter ** 0.4)) * 
                 (GlobeTC - AirTC)) ** (1 / 4) - ConstantsLib.T_kel
        return T_mrt
